from typing import Any, Dict, List, Union

import pandas as pd

from helperpy.core.string_ops import remove_last_n_characters
from helperpy.core.utils import is_none_or_nan
from helperpy.data_wrangler.explore import get_column_availability_info


def wrap_string_with_appropriate_quotes(text: str) -> str:
    """Wraps given string with appropriate quotes (for SQL insert query)"""
    has_single_quote = "'" in text
    has_double_quote = '"' in text
    if not has_single_quote and not has_double_quote:
        return f"'{text}'"
    if has_single_quote and not has_double_quote:
        return f'"{text}"'
    if not has_single_quote and has_double_quote:
        return f"'{text}'"
    return f'"{text}"'


def get_in_query_for_numbers(numbers: List[Union[int, float]]) -> str:
    comma_separated_numbers = ", ".join(map(str, numbers))
    return f"({comma_separated_numbers})"


def get_in_query_for_strings(strings: List[str]) -> str:
    comma_separated_strings = ", ".join(map(lambda string: wrap_string_with_appropriate_quotes(text=string), strings))
    return f"({comma_separated_strings})"


def __wrap_boolean_value(value: Any) -> str:
    """
    If `value` is Python's native True, returns 'true'.
    If `value` is Python's native False, returns 'false'.
    Otherwise returns 'null'.
    """
    if value is True:
        return 'true'
    if value is False:
        return 'false'
    return 'null'


def __wrap_value_by_sql_datatype(
        value: Union[int, float, str, None],
        datatype: str,
    ) -> Union[int, float, str]:
    # Handle nulls
    if is_none_or_nan(value=value):
        return 'null'
    
    # Handle non-nulls
    if datatype == 'integer':
        return int(value)
    elif datatype == 'float':
        return float(value)
    elif datatype == 'string':
        return wrap_string_with_appropriate_quotes(text=str(value))
    elif datatype == 'date':
        return f"DATE('{value}')"
    elif datatype == 'timestamp':
        return f"TIMESTAMP('{value}')"
    elif datatype == 'boolean':
        return __wrap_boolean_value(value=value)
    raise ValueError(f"Got unexpected option for `datatype`: '{datatype}'")


def __get_query_string_of_records(
        records: List[Dict],
        columns: List[str],
        datatypes: List[str],
    ) -> str:
    query = ""
    for record in records:
        values_per_record = []
        for column, datatype in zip(columns, datatypes):
            value = __wrap_value_by_sql_datatype(
                value=record[column],
                datatype=datatype,
            )
            values_per_record.append(value)
        comma_separated_values = ', '.join(map(str, values_per_record))
        query += f"({comma_separated_values}),\n"
    query = remove_last_n_characters(text=query, num_chars=2)
    return query


def generate_insert_query(
        data: pd.DataFrame,
        table_name: str,
        column_to_datatype_mapper: Dict[str, str],
    ) -> str:
    """
    Takes in DataFrame having the data to be inserted into a table.
    Returns an INSERT query (string) for the given DataFrame.
    Evaluates Python's None and Numpy's NaN to be null values.
    Expects column names of the DataFrame to match the column names of the table
    where the records will be inserted.

    Accepted data-type options for the columns in the DataFrame:
        - integer
        - float
        - string
        - date: Values must be strings of format "yyyy-mm-dd". Eg: "2020-03-15"
        - timestamp: Values must be strings of format: "yyyy-mm-dd hh:mm:ss". Eg: "2020-03-15 19:30:55"
        - boolean: Values must be Python's native True/False. Anything else will be considered as a null value.
    
    >>> generate_insert_query(
            data=data, # Any DataFrame
            table_name='employee',
            column_to_datatype_mapper={
                'name': 'string',
                'age': 'integer',
                'date_of_birth': 'date',
                'joined_at': 'timestamp',
                'salary': 'float',
                'is_recent_recruit': 'boolean',
            },
        )
    """
    records = data.to_dict(orient='records')
    columns = list(column_to_datatype_mapper.keys())
    datatypes = list(column_to_datatype_mapper.values())

    # Data validation
    columns_missing = get_column_availability_info(data=data, expected_columns=columns)['columns_missing']
    if columns_missing:
        raise KeyError(f"The following columns from `column_to_datatype_mapper` are missing in the given DataFrame: {columns_missing}")
    valid_datatype_options = ['integer', 'float', 'string', 'date', 'timestamp', 'boolean']
    invalid_datatypes = list(set(datatypes).difference(set(valid_datatype_options)))
    if invalid_datatypes:
        raise ValueError(f"The following datatypes from `column_to_datatype_mapper` are invalid: {invalid_datatypes}. Valid datatype options are: {valid_datatype_options}")
    
    # Actual logic
    comma_separated_column_names = ', '.join(map(str, columns))
    query_string_of_records = __get_query_string_of_records(records=records, columns=columns, datatypes=datatypes)
    query = f"""INSERT INTO {table_name} ({comma_separated_column_names})\nVALUES\n{query_string_of_records}"""
    return query


import logging


def get_logger_object(
        logger_name: str,
        filepath: str,
    ) -> logging.Logger:
    """
    Gets `logging.Logger` object based on the given params. Each set of params must be unique together.
    Can be used to write neat-looking log messages for all possible log levels.

    >>> logger_ceo = get_logger_object(logger_name="ceo", filepath="ceo.log")
    >>> logger_emp = get_logger_object(logger_name="employee", filepath="employee.log")
    >>> logger_ceo.info(msg="This is a log in the CEO module")
    >>> logger_emp.info(msg="This is a log in the Employee module")
    
    ### References
        - https://docs.python.org/3/library/logging.html
        - https://docs.python.org/3/library/logging.html#logrecord-attributes
        - https://docs.python.org/3/howto/logging-cookbook.html
        - https://www.youtube.com/watch?v=jxmzY9soFXg
    """
    logger_obj = logging.getLogger(name=logger_name)
    logger_obj.setLevel(level=logging.DEBUG)
    file_handler = logging.FileHandler(
        filename=filepath,
        mode='a',
    )
    file_handler.setFormatter(
        fmt=logging.Formatter(
            fmt="\nDatetime: %(asctime)s | Level: %(levelname)s | Logger: %(name)s | File: %(filename)s | Function/method: %(funcName)s | Message: %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S",
        )
    )
    logger_obj.addHandler(hdlr=file_handler)
    return logger_obj


from typing import List, Optional
from email import encoders
from email.mime.audio import MIMEAudio
from email.mime.base import MIMEBase
from email.mime.image import MIMEImage
from email.mime.multipart import MIMEMultipart
from email.mime.text import MIMEText
import mimetypes
import smtplib

from helperpy.core.file_io import get_basename_from_filepath


def __add_attachments_to_multipart_object(
        filepaths_to_attachments: List[str],
        multipart_obj: MIMEMultipart,
    ) -> MIMEMultipart:
    """
    Adds attachments from the given filepaths to the given `MIMEMultipart` object.
    Returns `MIMEMultipart` object containing the attachments.

    Resources:
        - [Python email package examples](https://www.rose-hulman.edu/class/cs/archive/csse120-old/csse120-old-terms/201210/Resources/python-3.1.2-docs-html/library/email-examples.html)
        - [Common MIME types](https://developer.mozilla.org/en-US/docs/Web/HTTP/Basics_of_HTTP/MIME_types/Common_types)
    """
    for filepath in filepaths_to_attachments:
        type_, encoding = mimetypes.guess_type(url=filepath)
        if type_ is None or encoding is not None:
            type_ = 'application/octet-stream'
        maintype, subtype = type_.split(sep='/', maxsplit=1)
        if maintype == 'text':
            fp = open(file=filepath)
            payload = MIMEText(_text=fp.read(), _subtype=subtype)
            fp.close()
        elif maintype == 'image':
            fp = open(file=filepath, mode='rb')
            payload = MIMEImage(_imagedata=fp.read(), _subtype=subtype)
            fp.close()
        elif maintype == 'audio':
            fp = open(file=filepath, mode='rb')
            payload = MIMEAudio(_audiodata=fp.read(), _subtype=subtype)
            fp.close()
        else:
            fp = open(file=filepath, mode='rb')
            payload = MIMEBase(_maintype=maintype, _subtype=subtype)
            payload.set_payload(payload=fp.read())
            fp.close()
            encoders.encode_base64(msg=payload)
        payload.add_header(
            _name='Content-Disposition',
            _value='attachment',
            filename=get_basename_from_filepath(filepath=filepath),
        )
        multipart_obj.attach(payload=payload)
    return multipart_obj


def send_email(
        from_email_id: str,
        from_email_id_password: str,
        to_email_ids: List[str],
        cc_email_ids: List[str],
        subject: str,
        body: str,
        filepaths_to_attachments: Optional[List[str]] = None,
    ) -> None:
    """
    Sends an email from one Email ID to one or more Email IDs, along with the attachments provided (if any).
    Attachments work for the following file extensions: ['csv', 'docx', 'flv', 'jpg', 'm4a', 'mp3', 'mp4', 'pdf', 'png', 'txt', 'xls', 'xlsx', 'zip'].
    Accepts HTML tags for the `body` parameter.
    Returns None if the email is sent successfully; otherwise raises an Exception.

    >>> send_email(
            from_email_id='sender_email_id@gmail.com',
            from_email_id_password='some_password',
            to_email_ids=['person1@gmail.com', 'person2@gmail.com'],
            cc_email_ids=['person3@gmail.com'],
            subject="Your subject",
            body="Your message",
            filepaths_to_attachments=['file1.xlsx', 'file2.pdf'],
        )
    """
    msg = MIMEMultipart()
    msg['From'] = from_email_id
    msg['To'] = ", ".join(to_email_ids)
    msg['Cc'] = ", ".join(cc_email_ids)
    msg['Subject'] = subject
    msg.attach(payload=MIMEText(_text=body, _subtype='html'))
    if filepaths_to_attachments is not None:
        msg = __add_attachments_to_multipart_object(
            filepaths_to_attachments=filepaths_to_attachments,
            multipart_obj=msg,
        )
    server = smtplib.SMTP(host='smtp.gmail.com', port=587)
    server.starttls()
    server.ehlo()
    server.login(user=from_email_id, password=from_email_id_password)
    server.sendmail(
        from_addr=msg['From'],
        to_addrs=msg['To'].split(',') + msg['Cc'].split(','),
        msg=msg.as_string(),
    )
    server.quit()
    return None


from typing import Callable
import functools
import random
import time

from helperpy.core.utils import get_timetaken_fstring


def timer(func: Callable) -> Callable:
    """Decorator that prints the runtime of the decorated function"""
    @functools.wraps(func)
    def wrapper_timer(*args, **kwargs):
        start = time.time()
        result = func(*args, **kwargs)
        end = time.time()
        time_taken_in_secs = round(end - start, 3)
        timetaken_fstring = get_timetaken_fstring(num_seconds=time_taken_in_secs)
        print(f"Executed {func.__name__!r} in: {timetaken_fstring}")
        return result
    return wrapper_timer


def slow_down(func: Callable) -> Callable:
    """
    Decorator that slows down the execution of the decorated function.
    Slows down the execution by 10-20 minutes.
    """
    @functools.wraps(func)
    def wrapper_slow_down(*args, **kwargs):
        num_seconds_delayed = random.randint(600, 1200)
        time.sleep(num_seconds_delayed)
        return func(*args, **kwargs)
    return wrapper_slow_down


def repeat(num_times: int) -> Callable:
    """Decorator that executes the decorated function `num_times` times"""
    def repeat_decorator(func):
        @functools.wraps(func)
        def wrapper(*args, **kwargs):
            for _ in range(num_times):
                result = func(*args, **kwargs)
            return result
        return wrapper
    return repeat_decorator


from typing import Dict, List, Optional, Union

import pandas as pd

from helperpy.core.exceptions import InvalidDataFrameError
from helperpy.core.type_annotations import NumberOrString


def is_dataframe_full(data: pd.DataFrame) -> bool:
    """Returns True if there are no missing values in a non-empty DataFrame"""
    is_not_empty = (not data.empty)
    has_no_missing_values = (data.isnull().sum().sum() == 0)
    return (is_not_empty & has_no_missing_values)


def get_column_count(dataframes: List[pd.DataFrame]) -> pd.DataFrame:
    """
    Returns DataFrame having count of all columns present in the given DataFrames.
    The returned DataFrame will have the following columns: ['Column', 'Count'].
    """
    all_columns = []
    for df in dataframes:
        all_columns.extend(df.columns.tolist())
    df_column_count = pd.Series(data=all_columns).value_counts().rename('Count').reset_index().rename(mapper={'index': 'Column'}, axis=1)
    return df_column_count


def get_common_columns(dataframes: List[pd.DataFrame]) -> Union[List[NumberOrString], List]:
    """Returns list of the common columns present in all the given DataFrames"""
    num_dataframes = len(dataframes)
    df_column_count = get_column_count(dataframes=dataframes)
    common_columns = df_column_count[df_column_count['Count'] == num_dataframes]['Column'].tolist()
    return common_columns


def get_column_availability_info(
        data: pd.DataFrame,
        expected_columns: List[NumberOrString],
    ) -> Dict[str, List[str]]:
    """
    Takes in non-empty DataFrame, and list of columns expected to be in said DataFrame.
    Returns dictionary having 2 keys: ['columns_available', 'columns_missing'], wherein the value for
    each key will be a list of columns that are available/missing.
    """
    if data.empty:
        raise InvalidDataFrameError("Expected a non-empty DataFrame, but got an empty DataFrame")
    all_columns = data.columns.tolist()
    expected_columns = list(expected_columns)
    columns_available = list(
        set(expected_columns).intersection(set(all_columns))
    )
    columns_missing = list(
        set(expected_columns).difference(set(all_columns))
    )
    dict_obj = {
        'columns_available': columns_available,
        'columns_missing': columns_missing,
    }
    return dict_obj


def has_all_expected_columns(
        data: pd.DataFrame,
        expected_columns: List[NumberOrString],
    ) -> bool:
    """Returns True if all expected columns are available in given DataFrame; otherwise returns False"""
    dict_column_availability_info = get_column_availability_info(
        data=data,
        expected_columns=expected_columns,
    )
    return len(dict_column_availability_info['columns_missing']) == 0


def describe_missing_data(
        data: pd.DataFrame,
        show_all_columns: Optional[bool] = True,
    ) -> pd.DataFrame:
    """
    Returns DataFrame having information about missing values (if any) present in a non-empty DataFrame.
    The DataFrame returned will have the following columns: ['Column', 'NumMissingValues', 'PercentMissingValues', 'DataType'].
    """
    if data.empty:
        raise InvalidDataFrameError("Expected a non-empty DataFrame, but got an empty DataFrame")
    columns = data.columns.tolist()
    num_missing_values = data.isnull().sum().values
    percent_missing_values = pd.Series(data=num_missing_values * 100 / len(data)).apply(round, args=[2])
    datatypes = data.dtypes.values
    df_missing_data_info = pd.DataFrame(data={
        'Column': columns,
        'NumMissingValues': num_missing_values,
        'PercentMissingValues': percent_missing_values,
        'DataType': datatypes,
    })
    df_missing_data_info.sort_values(
        by=['PercentMissingValues', 'Column'],
        ascending=[False, True],
        ignore_index=True,
        inplace=True,
    )
    if not show_all_columns:
        df_missing_data_info = df_missing_data_info[df_missing_data_info['PercentMissingValues'] > 0]
    return df_missing_data_info


from typing import Any, List, Optional
import json
import ntpath
import os
import shutil

import joblib
import numpy as np
import pandas as pd
from tinytag import TinyTag
from tqdm import tqdm

from helperpy.core.utils import get_timetaken_fstring
from helperpy.data_wrangler.transform import rank_and_sort


def get_extension(filepath: str) -> str:
    return os.path.splitext(filepath)[-1][1:]


def get_basename_from_filepath(filepath: str) -> str:
    """Returns base-name of the file/folder from the given `filepath` (along with the extension, if any)"""
    head, tail = ntpath.split(p=filepath)
    return tail or ntpath.basename(head)


def get_absolute_filepath(filepath: str) -> str:
    """Returns absolute filepath of the file/folder from the given `filepath` (along with the extension, if any)"""
    absolute_filepath = os.path.realpath(path=filepath)
    return absolute_filepath


def get_line_count(filepath: str) -> int:
    """Returns count of number of lines in the given file"""
    num_lines = sum(1 for _ in open(file=filepath, encoding="utf8"))
    return num_lines


def filter_filepaths_by_extensions(
        filepaths: List[str],
        extensions: List[str],
    ) -> List[str]:
    """
    Filters given filepaths by the desired extensions.

    >>> filter_filepaths_by_extensions(
        filepaths=['one.js', 'two.py', 'three.css', 'four.go', 'five.html', 'six.py', 'seven.js'],
        extensions=['css', 'js'],
    )
    """
    extensions = list(
        map(lambda extension: extension.strip().lower(), extensions)
    )
    filepaths_needed = list(
        filter(lambda filepath: get_extension(filepath=filepath).strip().lower() in extensions, filepaths)
    )
    return filepaths_needed


def get_unique_extensions(filepaths: List[str]) -> List[str]:
    """Returns all unique extensions available in the list of filepaths given"""
    all_extensions = list(map(get_extension, filepaths))
    unique_extensions = sorted(list(set(all_extensions)))
    return unique_extensions


def __get_filepaths_at_first_level(src_dir: str) -> List[str]:
    folders_and_files_in_directory = os.listdir(src_dir)
    folders_and_files_in_directory = list(
        map(lambda folder_or_file: os.path.join(src_dir, folder_or_file), folders_and_files_in_directory)
    )
    files_in_directory = list(
        filter(lambda folder_or_file: os.path.isfile(folder_or_file), folders_and_files_in_directory)
    )
    return files_in_directory


def __get_filepaths_at_all_levels(src_dir: str) -> List[str]:
    filepaths = []
    for path, _, filenames in os.walk(src_dir):
        for filename in filenames:
            filepath = os.path.join(path, filename)
            filepaths.append(filepath)
    return filepaths


def get_filepaths(
        src_dir: str,
        depth: str,
        extensions: Optional[List[str]] = None,
    ) -> List[str]:
    """
    Gets list of all filepaths (of files) from source directory.
    
    Parameters:
        - src_dir (str): Filepath to the source directory. Can be absolute or relative.
        - depth (str): Options: ['all_levels', 'first_level'].
        Set to 'all_levels' if you want to get filepaths from all sub-directories (if any) in the given source directory.
        Set to 'first_level' if you want to get filepaths only from the first directory given.
        - extensions (list): List of extensions to filter the filepaths by (optional).
    
    >>> get_filepaths(
            src_dir="SOME_SOURCE_DIR",
            depth='all_levels',
            extensions=['csv', 'xlsx'],
        )
    """
    depth_options = ['all_levels', 'first_level']
    if depth not in depth_options:
        raise ValueError(f"Expected `depth` to be in {depth_options}, but got '{depth}'")
    
    if depth == 'all_levels':
        filepaths = __get_filepaths_at_all_levels(src_dir=src_dir)
    elif depth == 'first_level':
        filepaths = __get_filepaths_at_first_level(src_dir=src_dir)
    if extensions is not None:
        filepaths = filter_filepaths_by_extensions(filepaths=filepaths, extensions=extensions)
    return filepaths


def create_archive_file(
        src_dir: str,
        archive_format: str,
    ) -> None:
    """
    Creates archive file of the given source directory.
    Options for `archive_format` are: ['zip', 'tar', 'gztar', 'bztar', 'xztar'].
    """
    absolute_path_to_src_dir = get_absolute_filepath(filepath=src_dir)
    basename_of_src_dir = get_basename_from_filepath(filepath=src_dir)
    shutil.make_archive(
        base_name=basename_of_src_dir,
        format=archive_format,
        root_dir=absolute_path_to_src_dir,
    )
    return None


def get_line_count_info(
        filepaths: List[str],
        rank_by_line_count: Optional[bool] = False,
    ) -> pd.DataFrame:
    """
    Gets line-count of all filepaths given.
    Returns DataFrame having columns: ['Filepath', 'LineCount', 'Basename', 'Extension'].
    If `rank_by_line_count` is set to True, then there will be an additional column called: 'RankByLineCount'.

    >>> get_line_count_info(
            filepaths=['file1.txt', 'file2.txt', 'file3.txt'],
            rank_by_line_count=True,
        )
    """
    df = pd.DataFrame()
    df['Filepath'] = filepaths
    df['LineCount'] = df['Filepath'].apply(get_line_count)
    df['Basename'] = df['Filepath'].apply(get_basename_from_filepath)
    df['Extension'] = df['Filepath'].apply(get_extension)
    if rank_by_line_count:
        df = rank_and_sort(
            data=df,
            rank_column_name='RankByLineCount',
            rank_by=['LineCount'],
            ascending=[False],
            how='dense_rank',
        )
    return df


def get_media_metadata(filepaths: List[str]) -> pd.DataFrame:
    """
    Gets metadata of all media-file filepaths given i.e; files with extensions mp3, mp4, etc.
    Returns DataFrame having columns: ['Filepath', 'Basename', 'Extension', 'BitRate',
    'DurationInSeconds', 'Duration', 'SizeInMb', 'SampleRate'].
    
    >>> get_media_metadata(filepaths=['audio.mp3', 'audio.m4a', 'video.mp4'])
    """
    bytes_per_mb = 1048576
    df = pd.DataFrame()
    df['TinyTagObject'] = [TinyTag.get(filepath) for filepath in tqdm(filepaths)]
    df['Filepath'] = filepaths
    df['Basename'] = df['Filepath'].apply(get_basename_from_filepath)
    df['Extension'] = df['Filepath'].apply(get_extension)
    df['BitRate'] = df['TinyTagObject'].apply(lambda obj: obj.bitrate)
    df['DurationInSeconds'] = df['TinyTagObject'].apply(
        lambda obj: None if obj.duration is None else int(np.ceil(obj.duration))
    )
    df['Duration'] = df['TinyTagObject'].apply(
        lambda obj: None if obj.duration is None else get_timetaken_fstring(
            num_seconds=int(np.ceil(obj.duration))
        )
    )
    df['SizeInMb'] = df['TinyTagObject'].apply(
        lambda obj: None if obj.filesize is None else round(obj.filesize / bytes_per_mb, 2)
    )
    df['SampleRate'] = df['TinyTagObject'].apply(lambda obj: obj.samplerate)
    df.drop(labels=['TinyTagObject'], axis=1, inplace=True)
    return df


def pickle_load(filepath: str) -> Any:
    """Loads data from pickle file, via joblib module"""
    python_obj = joblib.load(filename=filepath)
    return python_obj


def pickle_save(data_obj: Any, filepath: str) -> None:
    """Saves data as pickle file, via joblib module"""
    joblib.dump(value=data_obj, filename=filepath)
    return None


def save_object_as_json(obj: Any, filepath: str) -> None:
    """Saves Python object as JSON file"""
    with open(file=filepath, mode='w') as fp:
        json.dump(obj=obj, fp=fp, indent=4)
    return None


def read_to_dataframe(filepath: str) -> pd.DataFrame:
    """
    Reads DataFrame from filepaths having any of the following extensions: ['csv', 'xlsx'].
    >>> df = read_to_dataframe(filepath="some_data.csv")
    """
    extension = get_extension(filepath=filepath).lower()
    if extension == 'csv':
        data = pd.read_csv(filepath)
    elif extension == 'xlsx':
        data = pd.read_excel(filepath)
    else:
        raise ValueError(f"Expected filepath's extension to be in ['csv', 'xlsx'], but got '{extension}'")
    return data


def save_dataframe(
        data: pd.DataFrame,
        filepath: str,
    ) -> None:
    """
    Saves DataFrame to one of the following file types: ['csv', 'xlsx', 'json'].
    >>> save_dataframe(
            data=pd.DataFrame(data={'Column1': [1, 2, 3, 4]}), # Some DataFrame
            filepath="some_data.csv",
        )
    """
    extension = get_extension(filepath=filepath).lower()
    if extension == 'csv':
        data.to_csv(filepath, index=False)
    elif extension == 'xlsx':
        data.to_excel(filepath, index=False)
    elif extension == 'json':
        data.to_json(filepath, orient='records', indent=4)
    else:
        raise ValueError(f"Expected filepath's extension to be in ['csv', 'xlsx', 'json'], but got '{extension}'")
    return None


from typing import Any, Callable, Dict, List, Optional, Union
from collections import Counter
from functools import reduce
import random

import numpy as np
import pandas as pd

from helperpy.core.date_ops import get_current_timestamp_as_integer
from helperpy.core.text_casing import (
    camel_to_pascal,
    camel_to_snake,
    pascal_to_camel,
    pascal_to_snake,
    snake_to_camel,
    snake_to_pascal,
)
from helperpy.core.type_annotations import (
    Number,
    NumberOrString,
)
from helperpy.core.utils import Partitioner


def get_mapping_between_columns(
        data: pd.DataFrame,
        from_: NumberOrString,
        to: NumberOrString,
    ) -> Dict:
    """Returns dictionary having mappings between the `from_` and `to` columns of the given DataFrame"""
    from_column_has_nulls = (data[from_].isnull().sum() > 0)
    from_column_has_unique_values = (data[from_].nunique() == len(data))
    if from_column_has_nulls:
        raise ValueError("Values in the `from_` column must be non-nulls")
    if not from_column_has_unique_values:
        raise ValueError("Values in the `from_` column must be unique, as they will be the keys of the returned dictionary")
    mapper_dictionary = data.loc[:, [from_, to]].set_index(keys=[from_]).to_dict()[to]
    return mapper_dictionary


def spread_array_by_factor(
        array: List[Number],
        factor: int,
    ) -> List[Number]:
    """
    Spread out an array by given factor.
    >>> spread_array_by_factor(array=[4, 6, 7, 3], factor=3)
    >>> [4, 4, 4, 6, 6, 6, 7, 7, 7, 3, 3, 3]
    """
    array_after_spreading = []
    for idx, _ in enumerate(array):
        array_after_spreading += [array[idx]] * int(factor)
    return array_after_spreading


def spread_array_by_length(
        array: List[Number],
        to: int,
    ) -> List[Number]:
    """
    Definition:
        Spread out an array to particular length without distorting signal of the original data.
    Parameters:
        - array (list): List or list-like array data
        - to (int): Length of array to be spread into
    >>> spread_array_by_length(array=[2, 3, -7], to=14)
    >>> [2, 2, 2, 2, 3, 3, 3, 3, 3, -7, -7, -7, -7, -7]
    """
    initial_array_length = len(array)
    if initial_array_length >= to:
        return array
    array_after_spread = []
    spread_factor = to / initial_array_length
    # If length of array to spread into is divisible by length of initial array
    if int(spread_factor) == spread_factor:
        array_after_spread = spread_array_by_factor(array=array, factor=int(spread_factor))
        return array_after_spread
    # Perform necessary complete fill-ups of the array to spread into
    num_complete_fillups = int(np.floor(spread_factor))
    if num_complete_fillups > 0:
        array_after_spread = spread_array_by_factor(array=array, factor=num_complete_fillups)
    # Fill-up the remaining elements of `array_after_spread` randomly (to minimise distortion)
    while len(array_after_spread) != to:
        random_index = random.randint(0, len(array_after_spread)-1)
        random_element = array_after_spread[random_index]
        array_after_spread.insert(random_index+1, random_element)
    return array_after_spread


def linspace_by_index(
        array: List[Any],
        how_many: int,
    ) -> List[Any]:
    """
    Gets linspaced values from array (based on indices of the array).
    >>> array = [0, 8, 16, 24, 32, 40, 48, 56, 64, 72, 80, 88, 96]
    >>> linspace_by_index(array=array, how_many=4) # Returns [0, 32, 64, 96]
    >>> linspace_by_index(array=array, how_many=5) # Returns [0, 24, 48, 72, 96]
    """
    linspaced_indices = list(np.linspace(start=0, stop=len(array) - 1, num=how_many))
    linspaced_indices = list(map(np.ceil, linspaced_indices))
    linspaced_indices = list(map(int, linspaced_indices))
    linspaced_values = [array[idx] for idx in linspaced_indices]
    return linspaced_values


def normalize_array(array: List[Number]) -> List[Number]:
    """
    Normalizes values in array to range of [0, 1].
    Ignores and removes all NaNs.
    """
    array = np.array(array)
    array = array[~np.isnan(array)]
    normalized_array = (array - np.min(array)) / np.ptp(array)
    return list(normalized_array)


def dataframe_to_list(data: pd.DataFrame) -> Union[List[Dict], List]:
    """Converts DataFrame to a list of dictionaries"""
    if data.empty:
        return []
    return data.to_dict(orient='records')


def stringify_columns(data: pd.DataFrame) -> pd.DataFrame:
    """Converts all column names of the given DataFrame to strings (in case some of them are int/float)"""
    df = data.copy(deep=True)
    df.columns = list(map(str, df.columns.tolist()))
    return df


def merge_dataframes(
        dataframes: List[pd.DataFrame],
        how: str,
        on: List[NumberOrString],
    ) -> pd.DataFrame:
    """
    Merges list of DataFrames given.
    Parameters:
        - dataframes (list): List of DataFrames to merge.
        - how (str): The type of merge. Options: ['left', 'right', 'outer', 'inner', 'cross']
        - on (list): List of columns to merge on.
    """
    df_merged = reduce(
        lambda df_left, df_right: pd.merge(
            left=df_left,
            right=df_right,
            how=how,
            on=on,
        ),
        dataframes,
    )
    return df_merged


def drop_columns_if_exists(
        data: pd.DataFrame,
        columns: List[NumberOrString],
    ) -> pd.DataFrame:
    """Drops list of given columns if they exist in the given DataFrame"""
    df = data.copy(deep=True)
    columns_available = df.columns.tolist()
    columns_to_drop = list(
        set(columns_available).intersection(set(columns))
    )
    if columns_to_drop:
        df.drop(labels=columns_to_drop, axis=1, inplace=True)
    return df


def transform_datetime_columns(
        data: pd.DataFrame,
        func: Callable,
        subset: Optional[List[NumberOrString]] = None,
        column_prefix: Optional[str] = '',
        column_suffix: Optional[str] = '',
    ) -> pd.DataFrame:
    """
    Takes in a DataFrame, and applies the given function to all 'datetime64' columns in the DataFrame.
    Parameters:
        - data (DataFrame): Pandas DataFrame
        - func (callable): Callable Python function
        - subset (list): Subset of date/datetime columns to apply the function to (optional)
        - column_prefix (str): Prefix to add to column name, if you want to create new columns (optional)
        - column_suffix (str): Suffix to add to column name, if you want to create new columns (optional)
    """
    df = data.copy(deep=True)
    subset = df.select_dtypes(include=['datetime64']).columns.tolist() if subset is None else subset
    for column in subset:
        new_column = f"{column_prefix}{column}{column_suffix}"
        df[new_column] = df[column].apply(func=func)
    return df


def prettify_datetime_columns(
        data: pd.DataFrame,
        include_time: bool,
        subset: Optional[List[NumberOrString]] = None,
        column_prefix: Optional[str] = '',
        column_suffix: Optional[str] = '',
    ) -> pd.DataFrame:
    """
    Takes in Pandas DataFrame, and converts all 'datetime64' columns to a more human readable format.
    Parameters:
        - data (DataFrame): Pandas DataFrame
        - include_time (bool): Includes time element if set to True
        - subset (list): Subset of date/datetime columns to apply the function to (optional)
        - column_prefix (str): Prefix to add to column name, if you want to create new columns (optional)
        - column_suffix (str): Suffix to add to column name, if you want to create new columns (optional)
    """
    formatter = "%d %B, %Y %I:%M %p" if include_time else "%d %B, %Y"
    df = transform_datetime_columns(
        data=data,
        func=lambda dt_obj: dt_obj.strftime(formatter),
        subset=subset,
        column_prefix=column_prefix,
        column_suffix=column_suffix,
    )
    return df


def add_partitioning_column(
        data: pd.DataFrame,
        num_partitions: int,
        column_name: str,
    ) -> pd.DataFrame:
    """
    Partitions a DataFrame horizontally, based on number of partitions given.
    Returns DataFrame with an additional column containing the partition number.
    """
    df = data.copy(deep=True)
    partition_sizes = Partitioner(iterable_length=len(df)).sizes_by_num_partitions(
        num_partitions=num_partitions,
        distribution_method='uniform',
    )
    partition_column_values = []
    for idx, partition_size in enumerate(partition_sizes):
        partition_number = idx + 1
        partition_column_values.extend([partition_number] * partition_size)
    df[column_name] = partition_column_values
    return df


def partition_dataframe_by_num_partitions(
        data: pd.DataFrame,
        num_partitions: int,
    ) -> List[pd.DataFrame]:
    """
    Partitions a DataFrame horizontally, based on number of partitions given.
    Returns list of partitioned DataFrames.
    """
    df = data.copy(deep=True)
    partition_index_ranges = Partitioner(iterable_length=len(df)).index_ranges_by_num_partitions(
        num_partitions=num_partitions,
        distribution_method='uniform',
    )
    return [df.iloc[idx_start : idx_end] for idx_start, idx_end in partition_index_ranges]


def partition_dataframe_by_max_partition_length(
        data: pd.DataFrame,
        max_partition_length: int,
    ) -> List[pd.DataFrame]:
    """
    Partitions a DataFrame horizontally, based on maximum partition length given.
    Returns list of partitioned DataFrames.
    """
    num_partitions = int(np.ceil(len(data) / max_partition_length))
    return partition_dataframe_by_num_partitions(data=data, num_partitions=num_partitions)


def switch_column_casing(
        data: pd.DataFrame,
        casing_type: str,
    ) -> pd.DataFrame:
    """
    Switch casing of columns in DataFrame.
    Options for `casing_type`:
        * 'camel_to_pascal': Converts camel-case to pascal-case (Eg: someText --> SomeText)
        * 'camel_to_snake': Converts camel-case to snake-case (Eg: someText --> some_text)
        * 'pascal_to_camel': Converts pascal-case to camel-case (Eg: SomeText --> someText)
        * 'pascal_to_snake': Converts pascal-case to snake-case (Eg: SomeText --> some_text)
        * 'snake_to_camel': Converts snake-case to camel-case (Eg: some_text --> someText)
        * 'snake_to_pascal': Converts snake-case to pascal-case (Eg: some_text --> SomeText)
    
    Note: Expects all columns present in DataFrame to be of same casing.
    """
    df = data.copy(deep=True)
    mapper = {
        'camel_to_pascal': camel_to_pascal,
        'camel_to_snake': camel_to_snake,
        'pascal_to_camel': pascal_to_camel,
        'pascal_to_snake': pascal_to_snake,
        'snake_to_camel': snake_to_camel,
        'snake_to_pascal': snake_to_pascal,
    }
    columns = df.columns.tolist()
    df.columns = list(map(mapper[casing_type], columns))
    return df


def round_off_columns(
        data: pd.DataFrame,
        columns: List[NumberOrString],
        round_by: int,
    ) -> pd.DataFrame:
    """
    Rounds off specified numerical (float) columns in DataFrame.
    >>> round_off_columns(data=data, columns=['column1', 'column3', 'column5'], round_by=2)
    """
    df = data.copy(deep=True)
    for column in columns:
        df[column] = df[column].apply(round, args=[int(round_by)])
    return df


def __get_row_number_rankings(df_ranked: pd.DataFrame) -> List[int]:
    row_number_rankings = list(
        np.arange(start=1, stop=len(df_ranked) + 1, step=1)
    )
    return row_number_rankings


def __get_dense_rankings(
        df_ranked: pd.DataFrame,
        rank_by: List[NumberOrString],
    ) -> List[int]:
    dense_rankings = [1]
    values_used_for_ranking = list(df_ranked[rank_by].itertuples(index=False))
    prev = values_used_for_ranking[0]
    for value in values_used_for_ranking[1:]:
        latest_rank_assigned = dense_rankings[-1]
        ranking_for_row = latest_rank_assigned if prev == value else latest_rank_assigned + 1
        dense_rankings.append(ranking_for_row)
        prev = value
    return dense_rankings


def __get_non_dense_rankings(
        df_ranked: pd.DataFrame,
        rank_by: List[NumberOrString],
    ) -> List[int]:
    non_dense_rankings = [1]
    values_used_for_ranking = list(df_ranked[rank_by].itertuples(index=False))
    prev = values_used_for_ranking[0]
    for value in values_used_for_ranking[1:]:
        latest_rank_assigned = non_dense_rankings[-1]
        ranking_for_row = latest_rank_assigned if prev == value else latest_rank_assigned + Counter(non_dense_rankings)[latest_rank_assigned]
        non_dense_rankings.append(ranking_for_row)
        prev = value
    return non_dense_rankings


def rank_and_sort(
        data: pd.DataFrame,
        rank_column_name: NumberOrString,
        rank_by: List[NumberOrString],
        ascending: List[bool],
        how: str,
    ) -> pd.DataFrame:
    """
    Adds ranking column and sorts records based on the `rank_by` column/s.

    Parameters:
        - data (DataFrame): Pandas DataFrame.
        - rank_column_name (int | float | str): Name of the ranking column (the column which will contain the actual ranking).
        - rank_by (list): List of columns to rank by.
        - ascending (list): List of booleans signifying the order of ranking (must correspond to the columns in `rank_by`).
        - how (str): How the ranking should be implemented. Options: ['row_number', 'dense_rank', 'non_dense_rank'].
    
    |    | column   |   row_number |   dense_rank |   non_dense_rank |
    |---:|:---------|-------------:|-------------:|-----------------:|
    |  0 | a        |            1 |            1 |                1 |
    |  1 | a        |            2 |            1 |                1 |
    |  2 | a        |            3 |            1 |                1 |
    |  3 | a        |            4 |            1 |                1 |
    |  4 | b        |            5 |            2 |                5 |
    |  5 | b        |            6 |            2 |                5 |
    |  6 | c        |            7 |            3 |                7 |
    |  7 | d        |            8 |            4 |                8 |
    |  8 | d        |            9 |            4 |                8 |
    |  9 | e        |           10 |            5 |               10 |
    | 10 | e        |           11 |            5 |               10 |
    | 11 | f        |           12 |            6 |               12 |
    | 12 | g        |           13 |            7 |               13 |
    | 13 | g        |           14 |            7 |               13 |
    """
    how_options = ['row_number', 'dense_rank', 'non_dense_rank']
    if how not in how_options:
        raise ValueError(f"Expected `how` to be in {how_options}, but got '{how}'")
    if len(rank_by) != len(ascending):
        raise ValueError(
            "Expected `rank_by` and `ascending` to be of same length,"
            f" but got lengths {len(rank_by)} and {len(ascending)} respectively."
        )
    df_ranked = data.sort_values(by=rank_by, ascending=ascending, ignore_index=True).copy(deep=True)
    if how == 'row_number':
        rankings = __get_row_number_rankings(df_ranked=df_ranked)
    elif how == 'dense_rank':
        rankings = __get_dense_rankings(df_ranked=df_ranked, rank_by=rank_by)
    elif how == 'non_dense_rank':
        rankings = __get_non_dense_rankings(df_ranked=df_ranked, rank_by=rank_by)
    df_ranked[rank_column_name] = rankings
    column_order = [rank_column_name] + df_ranked.drop(labels=[rank_column_name], axis=1).columns.tolist()
    df_ranked = df_ranked.loc[:, column_order]
    return df_ranked


def normalize_numerical_columns(
        data: pd.DataFrame,
        columns: List[NumberOrString],
    ) -> pd.DataFrame:
    """
    Takes in DataFrame and list of numerical columns to normalize.
    Normalizes the given columns between [0-100].
    Note: Does not work with NaN values.
    """
    df = data.copy(deep=True)
    for column in columns:
        values = df[column].tolist()
        df[column] = normalize_array(array=values)
        df[column] = df[column].mul(100).round(2)
    return df


def randomly_fill_categorical_nans(
        data: pd.DataFrame,
        subset: Optional[List[NumberOrString]] = None,
    ) -> pd.DataFrame:
    """
    Fills missing values of categorical columns by selecting random value from said column.
    Parameters:
        - data (DataFrame): Pandas DataFrame
        - subset (list): Subset of categorical columns for which you want to randomly fill missing values (optional)
    """
    df = data.copy(deep=True)
    categorical_variables = df.select_dtypes(include='object').columns.tolist()
    if subset:
        categorical_variables = list(set(categorical_variables).intersection(set(subset)))
    null_indicator_string = f"NULL-{get_current_timestamp_as_integer()}-{random.randint(1000, 9999)}"
    for cv in categorical_variables:
        if df[cv].isnull().sum() > 0:
            df[cv].fillna(value=null_indicator_string, inplace=True)
            unique_choices = df[cv].unique().tolist()
            unique_choices.remove(null_indicator_string)
            df[cv] = df[cv].apply(
                lambda string: random.choice(unique_choices) if ((string == null_indicator_string) and unique_choices) else string
            )
    return df


from typing import Callable, Optional
import functools
import traceback

from django.db import (
    connection,
    reset_queries,
)
from rest_framework import status
from rest_framework.response import Response

from helperpy.frameworks.django_ops.utils import get_django_request_info_from_args


def django_query_counter(func: Callable) -> Callable:
    """Decorator that prints the number of Django queries executed by the decorated function"""
    @functools.wraps(func)
    def wrapper_django_query_counter(*args, **kwargs):
        reset_queries()
        num_queries_at_start = len(connection.queries)
        result = func(*args, **kwargs)
        num_queries_at_end = len(connection.queries)
        num_queries = num_queries_at_end - num_queries_at_start
        print(
            f"Number of Django queries executed by function {func.__name__!r} is: {num_queries}"
        )
        return result
    return wrapper_django_query_counter


def api_endpoint_exception_handler(
        api_endpoint_description: str,
        print_response: Optional[bool] = False,
    ) -> Callable:
    """
    Decorator that handles exceptions that may occur in function based API endpoints.
    Expects a function that takes in a `rest_framework.request.Request` object, and
    returns a `rest_framework.response.Response` object.
    If an exception is encountered, then a response with status code 500 is returned.
    """
    def decorator(func: Callable) -> Callable:
        @functools.wraps(func)
        def wrapper(*args, **kwargs) -> Response:
            try:
                response_obj = func(*args, **kwargs)
            except Exception as exc:
                response = {
                    'exception_type': type(exc).__name__,
                    'exception_msg': ' | '.join(exc.args),
                    'api_endpoint_function_name': func.__name__,
                    'api_endpoint_description': api_endpoint_description,
                    'traceback_string': traceback.format_exc(),
                    'request_info': get_django_request_info_from_args(*args),
                }
                response_obj = Response(data=response, status=status.HTTP_500_INTERNAL_SERVER_ERROR)
            if print_response:
                print(f"Response(data={response_obj.data}, status={response_obj.status_code})")
            return response_obj
        return wrapper
    return decorator


from typing import Any, Dict, List, Optional, Tuple, Union

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from helperpy.core.type_annotations import (
    Number,
    NumberOrString,
)
from helperpy.core.random_data_generator import generate_random_hex_codes
from helperpy.core.utils import (
    get_max_of_abs_values,
    has_negative_number,
    has_positive_number,
)
from helperpy.data_wrangler.transform import linspace_by_index


def __reorder_timseries_data(
        x_vals: List[NumberOrString],
        y_vals: List[Number],
    ) -> pd.DataFrame:
    """
    Re-orders values of x-axis and y-axis (for timeseries chart data) in ascending order of x-axis' values.
    Returns DataFrame of re-ordered values. Columns returned are: `['x_vals', 'y_vals']`
    """
    if len(x_vals) != len(y_vals):
        raise ValueError(
            "Expected `x_vals` and `y_vals` to be of same length, but"
            f" got lengths {len(x_vals)} and {len(y_vals)} respectively"
        )
    data = pd.DataFrame(data={
        'x_vals': x_vals,
        'y_vals': y_vals,
    })
    data.sort_values(by='x_vals', ascending=True, ignore_index=True, inplace=True)
    return data


def __plot_descriptive_stats(
        array: List[Number],
        orient: str,
        line_style: Optional[str] = "-.",
        line_width: Optional[Number] = 2,
    ) -> Any:
    """
    Helper function to plot descriptive stats along one of the axes. Options for `orient`: ['v', 'h'].
    Returns matplotlib plot object that can be chained to more matplotlib code.
    """
    mean = round(np.mean(array), 3)
    median = round(np.median(array), 3)
    if orient == 'v':
        obj = plt.axvline(x=mean, label=f"Mean ({mean})", linestyle=line_style, linewidth=line_width, color='red')
        obj = plt.axvline(x=median, label=f"Median ({median})", linestyle=line_style, linewidth=line_width, color='green')
    elif orient == 'h':
        obj = plt.axhline(y=mean, label=f"Mean ({mean})", linestyle=line_style, linewidth=line_width, color='red')
        obj = plt.axhline(y=median, label=f"Median ({median})", linestyle=line_style, linewidth=line_width, color='green')
    else:
        raise ValueError(f"Expected `orient` to be in ['v', 'h'], but got '{orient}'")
    return obj


def get_calibrated_plot_fonts(fig_size: Tuple[int, int]) -> Dict[str, float]:
    """
    Takes in figure-size (tuple), and returns dictionary containing fontsizes for all other aspects
    of the plot (calculated and set appropriately, based on the figure-size).
    The fontsizes dictionary will have the following keys: `['title_size', 'label_size', 'tick_size', 'legend_label_size', 'annotation_size']`
    """
    if isinstance(fig_size, tuple):
        if len(fig_size) == 2:
            width = int(fig_size[0])
        else:
            raise ValueError (f"Invalid value for `fig_size`. Expected tuple of length 2, but got tuple of length {len(fig_size)}")
    else:
        raise TypeError(f"Invalid type for `fig_size`. Expected 'tuple' but got '{type(fig_size)}'")
    title_to_width_ratio = round((40 / 24), 3)
    labels_to_width_ratio = round((25 / 24), 3)
    ticks_to_width_ratio = round((20 / 24), 3)
    legend_labels_to_width_ratio = round((24 / 24), 3)
    annotation_to_width_ratio = round((18 / 24), 3)
    dictionary_calibrated_plot_fonts = {
        'title_size': width * title_to_width_ratio,
        'label_size': width * labels_to_width_ratio,
        'tick_size': width * ticks_to_width_ratio,
        'legend_label_size': width * legend_labels_to_width_ratio,
        'annotation_size': width * annotation_to_width_ratio,
    }
    return dictionary_calibrated_plot_fonts


def add_plot_skeleton(
        title: str,
        x_label: str,
        y_label: str,
        fig_size: Tuple[int, int],
        include_labels: Optional[bool] = True,
        include_ticks: Optional[bool] = True,
    ) -> Any:
    """Returns matplotlib plot object, containing skeleton of basic aspects of a plot"""
    dict_plot_fonts = get_calibrated_plot_fonts(fig_size=fig_size)
    obj = plt.figure(figsize=fig_size)
    obj = plt.title(title, fontsize=dict_plot_fonts['title_size'])
    if include_labels:
        obj = plt.xlabel(x_label, fontsize=dict_plot_fonts['label_size'])
        obj = plt.ylabel(y_label, fontsize=dict_plot_fonts['label_size'])
    if include_ticks:
        obj = plt.xticks(fontsize=dict_plot_fonts['tick_size'])
        obj = plt.yticks(fontsize=dict_plot_fonts['tick_size'])
    return obj


def plot_donut(
        title: str,
        labels: List[str],
        values: List[Number],
        colors: Optional[List[str]] = None,
        save_at: Optional[str] = None,
        show: Optional[bool] = False,
    ) -> None:
    if not colors:
        colors = generate_random_hex_codes(how_many=len(labels))
    _, ax1 = plt.subplots()
    ax1.pie(x=list(values), labels=list(labels), colors=colors, autopct='%1.1f%%', startangle=40)
    centre_circle = plt.Circle((0,0), 0.70, fc='white')
    fig = plt.gcf()
    fig.gca().add_artist(centre_circle)
    ax1.axis('equal')  
    plt.tight_layout()
    plt.title(title, fontsize=15)
    if save_at:
        plt.savefig(save_at)
    if show:
        plt.show()
    plt.close()
    return None


def plot_bar(
        title: str,
        x_label: str,
        y_label: str,
        horizontal: bool,
        bar_labels: List[str],
        bar_values: List[Number],
        fig_size: Optional[Tuple[int, int]] = (12, 5),
        colors: Optional[List[str]] = None,
        annotate: Optional[bool] = True,
        symmetrical: Optional[bool] = True,
        save_at: Optional[str] = None,
        show: Optional[bool] = False,
    ) -> None:
    add_plot_skeleton(title=title,
                      x_label=x_label,
                      y_label=y_label,
                      fig_size=fig_size,
                      include_labels=True,
                      include_ticks=True)
    if not colors:
        colors = generate_random_hex_codes(how_many=len(bar_labels))
    if horizontal:
        plt.barh(y=bar_labels, width=bar_values, color=colors)
    else:
        plt.bar(x=bar_labels, height=bar_values, color=colors)
    if annotate:
        dict_plot_fonts = get_calibrated_plot_fonts(fig_size=fig_size)
        annotation_size = dict_plot_fonts['annotation_size']
    if annotate and horizontal:
        for idx, value in enumerate(bar_values):
            plt.text(x = value * 1.01,
                     y = idx,
                     s = str(value),
                     fontweight='bold',
                     fontsize=annotation_size,
                     color='black')
    if annotate and not horizontal:
        for idx, value in enumerate(bar_values):
            plt.text(x = idx,
                     y = value * 1.01,
                     s = str(value),
                     fontweight='bold',
                     fontsize=annotation_size,
                     color='black')
    if symmetrical:
        if has_negative_number(array=bar_values) and has_positive_number(array=bar_values):
            axis_limit = get_max_of_abs_values(array=bar_values) * 1.05
            if horizontal:
                plt.xlim(-axis_limit, axis_limit)
            else:
                plt.ylim(-axis_limit, axis_limit)
    if save_at:
        plt.savefig(save_at)
    if show:
        plt.show()
    plt.close()
    return None


def plot_valuecounts(
        title: str,
        array: List[Any],
        array_description: str,
        fig_size: Optional[Tuple[int, int]] = (12, 5),
        normalize: Optional[bool] = False,
        colors: Optional[List[str]] = None,
        symmetrical: Optional[bool] = True,
        save_at: Optional[str] = None,
        show: Optional[bool] = False,
    ) -> None:
    series = pd.Series(data=array)
    if normalize:
        dict_valuecounts = series.value_counts(normalize=True).mul(100).apply(round, args=[2]).to_dict()
        x_label = "Percentage of occurences"
    else:
        dict_valuecounts = series.value_counts().to_dict()
        x_label = "Count of occurences"
    plot_bar(title=title,
             x_label=x_label,
             y_label=array_description,
             horizontal=True,
             bar_labels=list(dict_valuecounts.keys()),
             bar_values=list(dict_valuecounts.values()),
             fig_size=fig_size,
             colors=colors,
             annotate=False,
             symmetrical=symmetrical,
             save_at=save_at,
             show=show)
    return None


def plot_timeseries(
        title: str,
        x_label: str,
        y_label: str,
        x_vals: List[Any],
        y_vals: List[Number],
        trim_xticks: Optional[bool] = True,
        fig_size: Optional[Tuple[int, int]] = (12, 5),
        area_fill: Optional[bool] = False,
        area_color: Optional[str] = "skyblue",
        area_alpha: Optional[float] = 0.4,
        line_width: Optional[Number] = 3,
        line_color: Optional[str] = "#135FB6",
        show_scatter_points: Optional[bool] = False,
        scatter_points_size: Optional[int] = 250,
        describe: Optional[bool] = False,
        symmetrical: Optional[bool] = True,
        save_at: Optional[str] = None,
        show: Optional[bool] = False,
    ) -> None:
    """Note: Expects `x_vals` to denote time-related data, and `y_vals` to denote the timeseries curve/line itself"""
    add_plot_skeleton(title=title,
                      x_label=x_label,
                      y_label=y_label,
                      fig_size=fig_size,
                      include_labels=True,
                      include_ticks=False)
    x_vals, y_vals = list(x_vals), list(y_vals)
    df_timeseries_data = __reorder_timseries_data(x_vals=x_vals, y_vals=y_vals)
    if area_fill:
        plt.fill_between(df_timeseries_data['x_vals'],
                         df_timeseries_data['y_vals'],
                         color=area_color,
                         alpha=area_alpha)
    plt.plot(df_timeseries_data['x_vals'],
             df_timeseries_data['y_vals'],
             color=line_color,
             linewidth=line_width)
    xticks = df_timeseries_data['x_vals'].tolist()
    if trim_xticks:
        xticks = linspace_by_index(array=xticks, how_many=5)
    dict_plot_fonts = get_calibrated_plot_fonts(fig_size=fig_size)
    tick_size = dict_plot_fonts['tick_size']
    plt.xticks(xticks, fontsize=tick_size)
    plt.yticks(fontsize=tick_size)
    if symmetrical:
        y_vals = df_timeseries_data['y_vals'].tolist()
        if has_negative_number(array=y_vals) and has_positive_number(array=y_vals):
            axis_limit = get_max_of_abs_values(array=y_vals) * 1.05
            plt.ylim(-axis_limit, axis_limit)
    if show_scatter_points:
        plt.scatter(df_timeseries_data['x_vals'],
                    df_timeseries_data['y_vals'],
                    color="black",
                    s=scatter_points_size,
                    marker='*',
                    label=None)
    if describe:
        __plot_descriptive_stats(array=y_vals, orient='h', line_style="-.", line_width=line_width)
    if describe: # Showing legend
        dict_plot_fonts = get_calibrated_plot_fonts(fig_size=fig_size)
        legend_label_size = dict_plot_fonts['legend_label_size']
        plt.legend(loc='best', fontsize=legend_label_size)
    plt.grid()
    plt.tight_layout()
    if save_at:
        plt.savefig(save_at)
    if show:
        plt.show()
    plt.close()
    return None


def plot_scatter(
        title: str,
        x_label: str,
        y_label: str,
        x_vals: List[Number],
        y_vals: List[Number],
        fig_size: Optional[Tuple[int, int]] = (12, 5),
        color: Optional[str] = 'g',
        alpha: Optional[float] = None,
        size: Optional[int] = 80,
        grid: Optional[bool] = True,
        save_at: Optional[str] = None,
        show: Optional[bool] = False,
    ) -> None:
    add_plot_skeleton(title=title,
                      x_label=x_label,
                      y_label=y_label,
                      fig_size=fig_size,
                      include_labels=True,
                      include_ticks=True)
    plt.scatter(x=list(x_vals), y=list(y_vals), c=color, s=size, alpha=alpha)
    if grid:
        plt.grid()
    plt.tight_layout()
    if save_at:
        plt.savefig(save_at)
    if show:
        plt.show()
    plt.close()
    return None


def plot_histogram(
        title: str,
        array: List[Number],
        array_description: str,
        fig_size: Optional[Tuple[int, int]] = (12, 5),
        bins: Optional[int] = 10,
        color: Optional[str] = "#2B88CA",
        describe: Optional[bool] = False,
        line_width: Optional[Number] = 3,
        grid: Optional[bool] = True,
        save_at: Optional[str] = None,
        show: Optional[bool] = False,
    ) -> None:
    add_plot_skeleton(title=title,
                      x_label=array_description,
                      y_label="Frequency",
                      fig_size=fig_size,
                      include_labels=True,
                      include_ticks=True)
    plt.hist(x=array, bins=bins, color=color)
    if describe:
        __plot_descriptive_stats(array=array, orient='v', line_style="-.", line_width=line_width)
    if describe: # Showing legend
        dict_plot_fonts = get_calibrated_plot_fonts(fig_size=fig_size)
        legend_label_size = dict_plot_fonts['legend_label_size']
        plt.legend(loc='best', fontsize=legend_label_size)
    if grid:
        plt.grid()
    plt.tight_layout()
    if save_at:
        plt.savefig(save_at)
    if show:
        plt.show()
    plt.close()
    return None


def plot_radar(
        title: str,
        labels: List[str],
        values: List[Number],
        ticks: Optional[List[Number]] = None,
        tick_limit: Optional[Tuple[Number, Number]] = None,
        fig_size: Optional[Tuple[int, int]] = (15, 9),
        color: Optional[str] = None,
        save_at: Optional[str] = None,
        show: Optional[bool] = False,
    ) -> None:
    dict_calibrated_plot_fonts = get_calibrated_plot_fonts(fig_size=fig_size)
    angles = np.linspace(start=0, stop=2*np.pi, num=len(labels), endpoint=False)
    values = np.concatenate((values, [values[0]]))
    labels = np.concatenate((labels, [labels[0]]))
    angles = np.concatenate((angles, [angles[0]]))
    
    fig = plt.figure(figsize=fig_size)
    ax = fig.add_subplot(111, polar=True)
    ax.plot(angles, values, 'o-', linewidth=2.5, c=color)
    ax.fill(angles, values, alpha=0.35, c=color)
    ax.set_thetagrids(angles * 180 / np.pi, labels, size=dict_calibrated_plot_fonts['label_size'])
    ax.grid(True)
    if ticks:
        plt.yticks(ticks, size=dict_calibrated_plot_fonts['tick_size'])
    else:
        plt.yticks([])
    if tick_limit:
        plt.ylim(tick_limit)
    plt.title(title, size=dict_calibrated_plot_fonts['title_size'])
    if save_at:
        plt.savefig(save_at)
    if show:
        plt.show()
    plt.close()
    return None


from typing import Any, List, Type


class InvalidOptionError(Exception):
    """Error raised when an invalid option is passed in"""
    pass


class InvalidDataFrameError(Exception):
    """Error raised when an invalid DataFrame is encountered (based on certain expectations)"""
    pass


class MissingRequiredParameterError(Exception):
    """Error raised when a required parameter is missing"""
    pass


def raise_exception_if_invalid_option(
        option_name: str,
        option_value: Any,
        valid_option_values: List[Any],
    ) -> None:
    """Raises a ValueError if the `option_value` given is an invalid option; otherwise returns None"""
    if option_value not in valid_option_values:
        raise InvalidOptionError(f"Expected `{option_name}` to be in {valid_option_values}, but got `{option_value}`")
    return None


def raise_exception_if_invalid_type(
        parameter_name: str,
        parameter_value: Any,
        expected_type: Type,
    ) -> None:
    """Raises a TypeError if the `parameter_value` given is not of the type `expected_type`; otherwise returns None"""
    if not isinstance(parameter_value, expected_type):
        raise TypeError(f"Expected `{parameter_name}` to be of type `{expected_type}`, but got type `{type(parameter_value)}`")
    return None


import re

ALPHABETS_LOWER_CASED = list('abcdefghijklmnopqrstuvwxyz')
ALPHABETS_UPPER_CASED = list(map(str.upper, ALPHABETS_LOWER_CASED))
ALPHABETS = ALPHABETS_LOWER_CASED + ALPHABETS_UPPER_CASED


def camel_to_pascal(string: str) -> str:
    """
    Converts camel-case to pascal-case.
    >>> camel_to_pascal(string="helloAndGoodMorning") # Returns "HelloAndGoodMorning"
    """
    return string[0].upper() + string[1:]


def pascal_to_camel(string: str) -> str:
    """
    Converts pascal-case to camel-case.
    >>> pascal_to_camel(string="HelloAndGoodMorning") # Returns "helloAndGoodMorning"
    """
    return string[0].lower() + string[1:]


def pascal_to_snake(string: str) -> str:
    """
    Converts pascal-case to snake-case.
    >>> pascal_to_snake(string="HelloAndGoodMorning") # Returns "hello_and_good_morning"
    """
    words = re.findall(pattern="[A-Z][^A-Z]*", string=string)
    words_lower_cased = list(map(str.lower, words))
    return "_".join(words_lower_cased)


def camel_to_snake(string: str) -> str:
    """
    Converts camel-case to snake-case.
    >>> camel_to_snake(string="helloAndGoodMorning") # Returns "hello_and_good_morning"
    """
    string_in_pascal = camel_to_pascal(string=string)
    string_in_snake = pascal_to_snake(string=string_in_pascal)
    return string_in_snake


def snake_to_pascal(string: str) -> str:
    """
    Converts snake-case to pascal-case.
    >>> snake_to_pascal(string="hello_and_good_morning") # Returns "HelloAndGoodMorning"
    """
    words = string.split('_')
    words_capitalized = list(map(str.capitalize, words))
    return "".join(words_capitalized)


def snake_to_camel(string: str) -> str:
    """
    Converts snake-case to camel-case.
    >>> snake_to_camel(string="hello_and_good_morning") # Returns "helloAndGoodMorning"
    """
    string_in_pascal = snake_to_pascal(string=string)
    string_in_camel = pascal_to_camel(string=string_in_pascal)
    return string_in_camel


def retardify(string: str) -> str:
    """
    Converts given text to retardified text.
    >>> retardify(string="Hello, and good morning!") # Returns "hElLo, AnD gOoD mOrNiNg!"
    """
    string = string.lower().strip()
    counter = 0
    retardified_text = ""
    for character in string:
        if character in ALPHABETS:
            counter += 1
            if counter % 2 == 0:
                character = character.upper()
        retardified_text += character
    return retardified_text


from typing import Any, Dict, List, Optional, Union
import mimetypes

from django.db.models import QuerySet
from django.http import HttpResponse
import pandas as pd
from rest_framework.request import Request

from helperpy.core.exceptions import MissingRequiredParameterError
from helperpy.core.file_io import get_basename_from_filepath
from helperpy.data_wrangler.transform import (
    dataframe_to_list,
    drop_columns_if_exists,
)


def queryset_to_dataframe(
        qs: QuerySet,
        fields_to_drop: Optional[List[str]] = None,
    ) -> pd.DataFrame:
    df = pd.DataFrame(data=list(qs.values()))
    if fields_to_drop:
        df = drop_columns_if_exists(data=df, columns=fields_to_drop)
    return df


def queryset_to_list(
        qs: QuerySet,
        fields_to_drop: Optional[List[str]] = None,
    ) -> Union[List[Dict], List]:
    df = queryset_to_dataframe(qs=qs, fields_to_drop=fields_to_drop)
    list_of_dicts = dataframe_to_list(data=df)
    return list_of_dicts


def file_to_django_http_response(filepath: str) -> HttpResponse:
    """
    Converts the given file to Django HTTP Response object.
    Works for the following file extensions: ['csv', 'docx', 'flv', 'jpg', 'm4a', 'mp3', 'mp4', 'pdf', 'png', 'txt', 'xls', 'xlsx', 'zip'].
    """
    type_, encoding = mimetypes.guess_type(url=filepath)
    filename = get_basename_from_filepath(filepath=filepath)
    if type_ is None or encoding is not None:
        type_ = 'application/octet-stream'
    with open(file=filepath, mode='rb') as fp:
        http_response = HttpResponse(content=fp, content_type=type_)
    http_response['Content-Disposition'] = f'attachment; filename="{filename}"'
    http_response['X-Sendfile'] = filename
    return http_response


def get_params_from_request_payload(
        request: Request,
        required_params: Optional[List[str]] = None,
        optional_params: Optional[List[str]] = None,
    ) -> Dict[str, Union[Any, None]]:
    """
    Gets parameters from the request's payload (the request object must be of type `rest_framework.request.Request`).
    Raises an exception if any of the `required_params` are missing.
    Otherwise returns a dictionary having keys = parameter name, and values = parameter value.
    """
    required_params = [] if required_params is None else required_params
    optional_params = [] if optional_params is None else optional_params
    all_params = required_params + optional_params
    all_unique_params = list(set(all_params))
    if len(all_params) != len(all_unique_params):
        raise ValueError(
            f"The given parameters (`required_params` + `optional_params`) must be unique. All parameters received: {all_params}"
        )
    
    dict_parsed_payload = {}
    for required_param in required_params:
        dict_parsed_payload[required_param] = request.data.get(required_param, None)
        if dict_parsed_payload[required_param] is None:
            raise MissingRequiredParameterError(
                f"The parameter '{required_param}' is required in the payload, but is not passed in"
            )
    for optional_param in optional_params:
        dict_parsed_payload[optional_param] = request.data.get(optional_param, None)
    return dict_parsed_payload


def get_query_params_from_request(
        request: Request,
        required_params: Optional[List[str]] = None,
        optional_params: Optional[List[str]] = None,
    ) -> Dict[str, Union[str, None]]:
    """
    Gets query parameters from the request (the request object must be of type `rest_framework.request.Request`).
    Raises an exception if any of the `required_params` are missing.
    Otherwise returns a dictionary having keys = parameter name, and values = parameter value.
    """
    required_params = [] if required_params is None else required_params
    optional_params = [] if optional_params is None else optional_params
    all_params = required_params + optional_params
    all_unique_params = list(set(all_params))
    if len(all_params) != len(all_unique_params):
        raise ValueError(
            f"The given parameters (`required_params` + `optional_params`) must be unique. All parameters received: {all_params}"
        )
    
    dict_query_params = {}
    for required_param in required_params:
        dict_query_params[required_param] = request.query_params.get(required_param, None)
        if dict_query_params[required_param] is None:
            raise MissingRequiredParameterError(
                f"The query-parameter '{required_param}' is required, but is not passed in"
            )
    for optional_param in optional_params:
        dict_query_params[optional_param] = request.query_params.get(optional_param, None)
    return dict_query_params


def get_django_request_info(request: Request) -> Dict[str, Any]:
    """
    Takes in DRF request object of type `rest_framework.request.Request` and returns
    dictionary having certain information about the request.
    Keys returned: ['content_type', 'method', 'session_info', 'stream', 'uri_absolute', 'uri_with_query_params', 'uri_without_query_params']
    """
    request_info = {
        'content_type': request.content_type,
        'method': request.method,
        'session_info': dict(request.session),
        'stream': request.stream,
        'uri_absolute': request.build_absolute_uri(),
        'uri_with_query_params': request.get_full_path(),
        'uri_without_query_params': request.path,
    }
    return request_info


def get_django_request_info_from_args(*args) -> Dict[str, Any]:
    """
    Searches for DRF request object of type `rest_framework.request.Request` (if any) from the given arguments.
    If it exists, returns dictionary having certain information about the request. Otherwise, returns an empty dictionary.
    """
    dict_request_info = {}
    for arg in args:
        if isinstance(arg, Request):
            dict_request_info = get_django_request_info(request=arg)
            break
    return dict_request_info


from typing import List, Optional, Tuple, Union

from datetime import (
    date,
    datetime,
    timedelta,
)

import pandas as pd
from randomtimestamp import randomtimestamp

from helperpy.core.exceptions import (
    raise_exception_if_invalid_option,
    raise_exception_if_invalid_type,
)


DATE_STRING_FORMAT = "%Y-%m-%d" # Eg: "2020-03-19"
DATETIME_STRING_FORMAT = "%Y-%m-%d %H:%M:%S" # Eg: "2020-03-19 17:45:08"
TIMESTAMP_INTEGER_FORMAT = "%Y%m%d%H%M%S" # Will be converted to an integer after parsing


class TimeUnitConverter:
    """Class for time related unit conversions"""
    SECONDS_PER_MINUTE = 60
    SECONDS_PER_HOUR = SECONDS_PER_MINUTE * 60
    SECONDS_PER_DAY = SECONDS_PER_HOUR * 24
    SECONDS_PER_WEEK = SECONDS_PER_DAY * 7
    SECONDS_PER_NON_LEAP_YEAR = SECONDS_PER_DAY * 365
    SECONDS_PER_LEAP_YEAR = SECONDS_PER_DAY * 366

    MINUTES_PER_HOUR = 60
    MINUTES_PER_DAY = MINUTES_PER_HOUR * 24
    MINUTES_PER_WEEK = MINUTES_PER_DAY * 7
    MINUTES_PER_NON_LEAP_YEAR = MINUTES_PER_DAY * 365
    MINUTES_PER_LEAP_YEAR = MINUTES_PER_DAY * 366

    HOURS_PER_DAY = 24
    HOURS_PER_WEEK = HOURS_PER_DAY * 7
    HOURS_PER_NON_LEAP_YEAR = HOURS_PER_DAY * 365
    HOURS_PER_LEAP_YEAR = HOURS_PER_DAY * 366



def get_current_timestamp_as_integer() -> int:
    """Returns current timestamp as an integer (Format: yyyymmddhhmmss)"""
    dt_obj = datetime.now()
    ts_now = dt_obj.strftime(TIMESTAMP_INTEGER_FORMAT)
    return int(ts_now)


def get_random_timestamp_as_integer() -> int:
    """Returns random timestamp as an integer (Format: yyyymmddhhmmss)"""
    ts_random = randomtimestamp(
        start_year=1900,
        end_year=datetime.now().year - 1,
        pattern=TIMESTAMP_INTEGER_FORMAT,
    )
    return int(ts_random)


def get_random_timestamp(
        start_year: Optional[int] = 1800,
        end_year: Optional[int] = 2200,
    ) -> datetime:
    return datetime.strptime(randomtimestamp(start_year=start_year, end_year=end_year), "%d-%m-%Y %H:%M:%S")


def period_to_datetime(period_obj: pd._libs.tslibs.period.Period) -> datetime:
    """Converts Pandas period object to datetime object"""
    dt_obj = datetime(
        year=period_obj.year,
        month=period_obj.month,
        day=period_obj.day,
        hour=period_obj.hour,
        minute=period_obj.minute,
        second=period_obj.second,
    )
    return dt_obj


def convert_to_datetime(
        date_string: str,
        hour: int,
        minute: int,
        second: int,
    ) -> datetime:
    """
    Parameters:
        - date_string (str): Date string of format "yyyy-mm-dd"
        - hour (int): Ranges from 0 to 23, indicating hour of the day
        - minute (int): Ranges from 0 to 59, indicating minute of the hour
        - second (int): Ranges from 0 to 59, indicating second of the minute
    """
    dt_obj = parse_date_string(date_string=date_string).replace(
        hour=hour,
        minute=minute,
        second=second,
    )
    return dt_obj


def get_day_of_week(
        dt_obj: Union[datetime, date],
        shorten: Optional[bool] = False,
    ) -> str:
    """
    Returns the day of the week.
    Day of week options when `shorten` is set to False: ['Monday', 'Tuesday', 'Wednesday', 'Thursday', 'Friday', 'Saturday', 'Sunday'].
    Day of week options when `shorten` is set to True: ['Mon', 'Tue', 'Wed', 'Thu', 'Fri', 'Sat', 'Sun'].
    """
    if shorten:
        return dt_obj.strftime("%a")
    return dt_obj.strftime("%A")


def to_date_string(dt_obj: datetime) -> str:
    """Converts datetime object to date string of format 'yyyy-mm-dd'"""
    date_string = dt_obj.strftime(DATE_STRING_FORMAT)
    return date_string


def to_datetime_string(dt_obj: datetime) -> str:
    """Converts datetime object to datetime string of format 'yyyy-mm-dd hh:mm:ss'"""
    datetime_string = dt_obj.strftime(DATETIME_STRING_FORMAT)
    return datetime_string


def parse_date_string(date_string: str) -> datetime:
    """
    Parses date string of format 'yyyy-mm-dd' into datetime object.
    >>> parse_date_string(date_string="2020-03-28")
    """
    dt_obj = datetime.strptime(date_string, DATE_STRING_FORMAT)
    return dt_obj


def parse_datetime_string(datetime_string: str) -> datetime:
    """
    Parses datetime string of format 'yyyy-mm-dd hh:mm:ss' into datetime object.
    Note: The 'hour' in the `datetime_string` must be of 24 hour format (from 0-23).
    >>> parse_datetime_string(datetime_string="2020-03-28 17:53:04")
    """
    dt_obj = datetime.strptime(datetime_string, DATETIME_STRING_FORMAT)
    return dt_obj


def ist_to_utc(dt_obj: datetime) -> datetime:
    """Subtracts 5 hours and 30 minutes from datetime object"""
    return dt_obj - timedelta(hours=5, minutes=30)


def utc_to_ist(dt_obj: datetime) -> datetime:
    """Adds 5 hours and 30 minutes to datetime object"""
    return dt_obj + timedelta(hours=5, minutes=30)


def convert_to_naive_timezone(dt_obj: datetime) -> datetime:
    """Converts datetime object to datetime object with a naive timezone"""
    dt_obj_naive_tz = dt_obj.replace(tzinfo=None)
    return dt_obj_naive_tz


def get_dates_between(
        start_date: str,
        end_date: str,
        as_type: Optional[str] = 'string',
    ) -> Union[List[datetime], List[date], List[str]]:
    """
    Returns list of dates between the given `start_date` and `end_date`.
    Input date-string format must be 'yyyy-mm-dd'.
    Options for `as_type` are: ['date', 'datetime', 'string']. Default: 'string'.
    """
    raise_exception_if_invalid_option(
        option_name='as_type',
        option_value=as_type,
        valid_option_values=['date', 'datetime', 'string'],
    )
    period_objs = pd.date_range(start=start_date, end=end_date, freq='1D')
    func_mapper = {
        'date': lambda period_obj: period_to_datetime(period_obj=period_obj).date(),
        'datetime': lambda period_obj: period_to_datetime(period_obj=period_obj),
        'string': lambda period_obj: to_date_string(dt_obj=period_to_datetime(period_obj=period_obj)),
    }
    return list(map(func_mapper[as_type], period_objs))


def offset_between_dates(
        start_date: str,
        end_date: str,
        offset_in_seconds: int,
    ) -> List[datetime]:
    """
    Returns list of datetime objects separated by the given offset (between the given date-range).

    >>> offset_between_dates(
            start_date="2019-05-15",
            end_date="2019-05-16",
            offset_in_seconds=60*60*8, # Equivalent of 8 hours
        )

    References:
        - [Pandas date range](https://pandas.pydata.org/docs/reference/api/pandas.date_range.html)
        - [Timeseries offset aliases](https://pandas.pydata.org/docs/user_guide/timeseries.html#timeseries-offset-aliases)
    """
    raise_exception_if_invalid_type(
        parameter_name='offset_in_seconds', parameter_value=offset_in_seconds, expected_type=int,
    )
    if offset_in_seconds <= 0:
        raise ValueError(f"Expected `offset_in_seconds` to be > 0, but got {offset_in_seconds}")
    
    dt_end_date = parse_date_string(date_string=end_date)
    periods_objs = pd.period_range(
        start=start_date,
        end=to_date_string(dt_obj=dt_end_date + timedelta(days=1)),
        freq=f"{offset_in_seconds}S",
    )
    dt_objs = list(map(period_to_datetime, periods_objs))
    if dt_objs[-1] == dt_end_date + timedelta(days=1):
        dt_objs.pop() # Remove last item
    return dt_objs


def offset_between_datetimes(
        start_timestamp: str,
        end_timestamp: str,
        offset_in_seconds: int,
        include_start_timestamp: Optional[bool] = True,
        include_end_timestamp: Optional[bool] = True,
    ) -> List[datetime]:
    """
    Returns list of datetime objects separated by the given offset (between the given datetime-range).

    >>> offset_between_datetimes(
            start_timestamp="2019-05-17 13:30:00",
            end_timestamp="2019-05-17 22:30:00",
            offset_in_seconds=60*60*1, # Equivalent of 1 hour
        )

    References:
        - [Pandas date range](https://pandas.pydata.org/docs/reference/api/pandas.date_range.html)
        - [Timeseries offset aliases](https://pandas.pydata.org/docs/user_guide/timeseries.html#timeseries-offset-aliases)
    """
    raise_exception_if_invalid_type(
        parameter_name='offset_in_seconds', parameter_value=offset_in_seconds, expected_type=int,
    )
    if offset_in_seconds <= 0:
        raise ValueError(f"Expected `offset_in_seconds` to be > 0, but got {offset_in_seconds}")
    
    periods_objs = pd.period_range(start=start_timestamp, end=end_timestamp, freq=f"{offset_in_seconds}S")
    dt_objs = list(map(period_to_datetime, periods_objs))
    if not include_start_timestamp and dt_objs:
        if start_timestamp == to_datetime_string(dt_obj=dt_objs[0]):
            dt_objs.pop(0)
    if not include_end_timestamp and dt_objs:
        if end_timestamp == to_datetime_string(dt_obj=dt_objs[-1]):
            dt_objs.pop(-1)
    return dt_objs



class DateWiseBucketer:
    """
    Class that provides functionality to get date-wise buckets of date-ranges.
    
    >>> dwb = DateWiseBucketer(
            date_string="2020-01-01",
            num_days_per_bucket=8,
            num_buckets=3,
        )
    >>> dwb.get_buckets(backward=False) # Returns [('2020-01-01', '2020-01-08'), ('2020-01-09', '2020-01-16'), ('2020-01-17', '2020-01-24')]
    >>> dwb.get_buckets(backward=True) # Returns [('2019-12-09', '2019-12-16'), ('2019-12-17', '2019-12-24'), ('2019-12-25', '2020-01-01')]
    """
    
    def __init__(
            self,
            date_string: str,
            num_days_per_bucket: int,
            num_buckets: int,
        ) -> None:
        """
        Parameters:
            - date_string (str): Date-string of format 'yyyy-mm-dd'
            - num_days_per_bucket (int): Number of days per bucket (must be > 0)
            - num_buckets (int): Number of buckets (must be > 0)
        """
        self.__raise_exception_if_invalid_input(
            date_string=date_string,
            num_days_per_bucket=num_days_per_bucket,
            num_buckets=num_buckets,
        )
        self.date_string = date_string
        self.num_days_per_bucket = num_days_per_bucket
        self.num_buckets = num_buckets
    
    def __raise_exception_if_invalid_input(
            self,
            date_string: str,
            num_days_per_bucket: int,
            num_buckets: int,
        ) -> None:
        """Raises an exception if any of the given inputs are invalid; otherwise returns None"""
        # Type errors
        raise_exception_if_invalid_type(
            parameter_name='date_string', parameter_value=date_string, expected_type=str,
        )
        raise_exception_if_invalid_type(
            parameter_name='num_days_per_bucket', parameter_value=num_days_per_bucket, expected_type=int,
        )
        raise_exception_if_invalid_type(
            parameter_name='num_buckets', parameter_value=num_buckets, expected_type=int,
        )

        # Value errors
        try:
            _ = datetime.strptime(date_string, DATE_STRING_FORMAT)
        except ValueError:
            raise ValueError(f"Expected `date_string` to be a valid date string of format 'yyyy-mm-dd', but got '{date_string}'")
        if num_days_per_bucket <= 0:
            raise ValueError(f"Expected `num_days_per_bucket` to be > 0, but got {num_days_per_bucket}")
        if num_buckets <= 0:
            raise ValueError(f"Expected `num_buckets` to be > 0, but got {num_buckets}")
        return None
    
    def __str__(self) -> str:
        return f"DateWiseBucketer(date_string='{self.date_string}', num_days_per_bucket={self.num_days_per_bucket}, num_buckets={self.num_buckets})"
    
    def get_buckets(
            self,
            backward: Optional[bool] = False,
            as_type: Optional[str] = 'string',
        ) -> Union[List[Tuple[datetime, datetime]], List[Tuple[date, date]], List[Tuple[str, str]]]:
        """
        Returns list of tuples of (start_date, end_date) buckets. They will always be in ascending order.
        Options for `as_type` are: ['date', 'datetime', 'string']. Default: 'string'.
        """
        raise_exception_if_invalid_option(
            option_name='as_type',
            option_value=as_type,
            valid_option_values=['date', 'datetime', 'string'],
        )
        freq = f"-{self.num_days_per_bucket}D" if backward else f"{self.num_days_per_bucket}D"
        period_objs = pd.date_range(start=self.date_string, freq=freq, periods=self.num_buckets)
        if backward:
            dt_bucket_end_dates = sorted(list(map(period_to_datetime, period_objs)), reverse=False)
            dt_bucket_start_dates = list(map(lambda dt_obj: dt_obj - timedelta(days=self.num_days_per_bucket - 1), dt_bucket_end_dates))
        else:
            dt_bucket_start_dates = sorted(list(map(period_to_datetime, period_objs)), reverse=False)
        buckets = list(map(lambda dt_obj: (dt_obj, dt_obj + timedelta(days=self.num_days_per_bucket - 1)), dt_bucket_start_dates))
        if as_type == 'date':
            buckets = list(map(lambda bucket: (bucket[0].date(), bucket[1].date()), buckets))
        elif as_type == 'string':
            buckets = list(map(lambda bucket: (to_date_string(bucket[0]), to_date_string(bucket[1])), buckets))
        return buckets



class MonthWiseBucketer:
    """
    Class that provides functionality to get month-wise buckets of date-ranges.
    
    >>> mwb = MonthWiseBucketer(
            year=2020,
            month=1,
            num_buckets=3,
        )
    >>> mwb.get_buckets(backward=False) # Returns [('2020-01-01', '2020-01-31'), ('2020-02-01', '2020-02-29'), ('2020-03-01', '2020-03-31')]
    >>> mwb.get_buckets(backward=True) # Returns [('2019-11-01', '2019-11-30'), ('2019-12-01', '2019-12-31'), ('2020-01-01', '2020-01-31')]
    """

    def __init__(
            self,
            year: int,
            month: int,
            num_buckets: int,
        ) -> None:
        """
        Parameters:
            - year (int): Year as integer
            - month (int): Month as integer (Range: 1-12)
            - num_buckets (int): Number of buckets (must be > 0)
        """
        self.__raise_exception_if_invalid_input(
            year=year,
            month=month,
            num_buckets=num_buckets,
        )
        self.year = year
        self.month = month
        self.num_buckets = num_buckets
    
    def __raise_exception_if_invalid_input(
            self,
            year: int,
            month: int,
            num_buckets: int,
        ) -> None:
        """Raises an exception if any of the given inputs are invalid; otherwise returns None"""
        # Type errors
        raise_exception_if_invalid_type(
            parameter_name='year', parameter_value=year, expected_type=int,
        )
        raise_exception_if_invalid_type(
            parameter_name='month', parameter_value=month, expected_type=int,
        )
        raise_exception_if_invalid_type(
            parameter_name='num_buckets', parameter_value=num_buckets, expected_type=int,
        )
        
        # Value errors
        if year <= 0:
            raise ValueError(f"Expected `year` to be > 0, but got {year}")
        if not (1 <= month <= 12):
            raise ValueError(f"Expected `month` to be in range 1-12, but got {month}")
        if num_buckets <= 0:
            raise ValueError(f"Expected `num_buckets` to be > 0, but got {num_buckets}")
        return None
    
    def __str__(self) -> str:
        return f"MonthWiseBucketer(year={self.year}, month={self.month}, num_buckets={self.num_buckets})"
    
    def get_buckets(
            self,
            backward: Optional[bool] = False,
            as_type: Optional[str] = 'string',
        ) -> Union[List[Tuple[datetime, datetime]], List[Tuple[date, date]], List[Tuple[str, str]]]:
        """
        Returns list of tuples of (start_date, end_date) buckets. They will always be in ascending order.
        Options for `as_type` are: ['date', 'datetime', 'string']. Default: 'string'.
        """
        raise_exception_if_invalid_option(
            option_name='as_type',
            option_value=as_type,
            valid_option_values=['date', 'datetime', 'string'],
        )
        period_objs = pd.date_range(
            start=datetime(year=self.year, month=self.month, day=1).strftime(DATE_STRING_FORMAT),
            freq="-1M" if backward else "1M",
            periods=self.num_buckets,
        )
        dt_bucket_end_dates = sorted(list(map(period_to_datetime, period_objs)), reverse=False)
        buckets = list(map(lambda dt_obj: (dt_obj.replace(day=1), dt_obj), dt_bucket_end_dates))
        if as_type == 'date':
            buckets = list(map(lambda bucket: (bucket[0].date(), bucket[1].date()), buckets))
        elif as_type == 'string':
            buckets = list(map(lambda bucket: (to_date_string(bucket[0]), to_date_string(bucket[1])), buckets))
        return buckets



class SecondWiseBucketer:
    """
    Class that provides functionality to get second-wise buckets of datetime-ranges.
    
    >>> swb = SecondWiseBucketer(
            start_timestamp="2019-05-18 17:00:00",
            end_timestamp="2019-05-18 17:20:00",
            offset_in_seconds=60*5, # Equivalent of 5 minutes
        )
    >>> swb.get_buckets() # Returns [('2019-05-18 17:00:00', '2019-05-18 17:05:00'), ('2019-05-18 17:05:00', '2019-05-18 17:10:00'), ('2019-05-18 17:10:00', '2019-05-18 17:15:00'), ('2019-05-18 17:15:00', '2019-05-18 17:20:00')]
    """

    def __init__(
            self,
            start_timestamp: str,
            end_timestamp: str,
            offset_in_seconds: int,
        ) -> None:
        self.start_timestamp = start_timestamp
        self.end_timestamp = end_timestamp
        self.offset_in_seconds = offset_in_seconds
    
    def __str__(self) -> str:
        return f"SecondWiseBucketer(start_timestamp='{self.start_timestamp}', end_timestamp='{self.end_timestamp}', offset_in_seconds={self.offset_in_seconds})"
    
    def get_buckets(self, as_type: Optional[str] = 'string') -> Union[List[Tuple[datetime, datetime]], List[Tuple[str, str]]]:
        """
        Returns list of tuples of (start_timestamp, end_timestamp) buckets. They will always be in ascending order.
        Options for `as_type` are: ['datetime', 'string']. Default: 'string'.
        """
        raise_exception_if_invalid_option(
            option_name='as_type',
            option_value=as_type,
            valid_option_values=['datetime', 'string'],
        )
        timestamps = offset_between_datetimes(
            start_timestamp=self.start_timestamp,
            end_timestamp=self.end_timestamp,
            offset_in_seconds=self.offset_in_seconds,
            include_start_timestamp=True,
            include_end_timestamp=True,
        )
        if as_type == 'string':
            timestamps = list(map(to_datetime_string, timestamps))
        buckets = [(timestamps[idx], timestamps[idx + 1]) for idx in range(len(timestamps) - 1)]
        return buckets


from asyncio import (
    ensure_future,
    gather,
    get_event_loop,
)
from typing import Any, Callable, Dict, List, Optional

from aiohttp import ClientSession


class HTTP_METHOD_NAME:
    """Exposes class variables having the various HTTP method names"""
    GET = "GET"
    POST = "POST"
    PUT = "PUT"
    PATCH = "PATCH"
    DELETE = "DELETE"


def __choose_session_method(
        http_method: str,
        session_obj: ClientSession,
    ) -> Callable:
    """Returns callable that is used to make the asynchronous API requests"""
    if http_method == HTTP_METHOD_NAME.GET:
        return session_obj.get
    elif http_method == HTTP_METHOD_NAME.POST:
        return session_obj.post
    elif http_method == HTTP_METHOD_NAME.PUT:
        return session_obj.put
    elif http_method == HTTP_METHOD_NAME.PATCH:
        return session_obj.patch
    elif http_method == HTTP_METHOD_NAME.DELETE:
        return session_obj.delete
    raise ValueError(f"Got an invalid `http_method` '{http_method}'")


async def __make_api_call(
        session: ClientSession,
        http_method: str,
        url: str,
        successful_status_codes: List[int],
        **request_kwargs,
    ) -> Dict[str, Any]:
    """Returns dictionary having keys: ['data', 'error']"""
    method_to_call = __choose_session_method(
        http_method=http_method,
        session_obj=session,
    )
    async with method_to_call(url, **request_kwargs) as response:
        if response.status in successful_status_codes:
            data = await response.json()
            error = None
        else:
            data = None
            error = {
                "url": url,
                "status_code": response.status,
                "text": await response.text(),
            }
    return {"data": data, "error": error}


async def __make_api_calls(
        http_method: str,
        urls: List[str],
        successful_status_codes: List[int],
        data_items: Optional[List[Any]] = None,
        **request_kwargs,
    ) -> List[Dict[str, Any]]:
    """Returns list of dictionaries having keys ['data', 'error'], for each URL (API endpoint) called"""
    if data_items and (http_method in [HTTP_METHOD_NAME.GET, HTTP_METHOD_NAME.DELETE]):
        raise ValueError(f"Cannot pass in `data_items` if `http_method` is set to '{http_method}'")
    if data_items and (len(data_items) != len(urls)):
        raise ValueError(
            "Length of `data_items` must be equal to the length of `urls`,"
            " as they must correspond to each other"
        )

    results = []
    actions = []
    if not data_items:
        async with ClientSession() as session:
            for url in urls:
                future_obj = ensure_future(
                    __make_api_call(
                        session=session,
                        http_method=http_method,
                        url=url,
                        successful_status_codes=successful_status_codes,
                        **request_kwargs,
                    )
                )
                actions.append(future_obj)
            results = await gather(*actions)
        return results

    async with ClientSession() as session:
        for url, data_item in zip(urls, data_items):
            future_obj = ensure_future(
                __make_api_call(
                    session=session,
                    http_method=http_method,
                    url=url,
                    successful_status_codes=successful_status_codes,
                    data=data_item,
                    **request_kwargs,
                )
            )
            actions.append(future_obj)
        results = await gather(*actions)
    return results


def make_api_calls(
        http_method: str,
        urls: List[str],
        successful_status_codes: List[int],
        data_items: Optional[List[Any]] = None,
        **request_kwargs,
    ) -> List[Dict[str, Any]]:
    """
    Returns list of dictionaries having keys ['data', 'error'], for each URL (API endpoint) called.

    Parameters:
        - http_method (str): HTTP method
        - urls (list): List of URLs (API endpoints) to call
        - successful_status_codes (list): List of status codes that are considered successful for the requests made
        - data_items (list): List of data items. Each item must correspond to the URLs in `urls` parameter
        (pass in `data_items` only for POST, PUT, PATCH methods)
        - request_kwargs: Kwargs related to the actual requests made i.e; headers.

    Docs: https://docs.aiohttp.org/en/stable/client_reference.html

    Usage:
    >>> num_api_calls = 1200
    >>> results = make_api_calls(
            http_method=HTTP_METHOD_NAME.GET,
            urls=[f"https://pokeapi.co/api/v2/pokemon/{number}" for number in range(1, num_api_calls+1)],
            successful_status_codes=[200],
            headers={},
        )
    >>> results = make_api_calls(
            http_method=HTTP_METHOD_NAME.POST,
            urls=["https://reqres.in/api/users"] * num_api_calls,
            successful_status_codes=[200, 201],
            data_items=[
                {
                    "name": f"MyName{number}",
                    "age": number,
                    "fav_movie": f"FavMovieName{number}",
                } for number in range(1, num_api_calls+1)
            ],
            headers={},
        )
    """
    event_loop = get_event_loop()
    results = event_loop.run_until_complete(
        __make_api_calls(
            http_method=http_method,
            urls=urls,
            successful_status_codes=successful_status_codes,
            data_items=data_items,
            **request_kwargs,
        )
    )
    # event_loop.close()
    return results


import numpy as np


def is_prime(number: int) -> bool:
    """Returns True if the given integer is prime; otherwise returns False"""
    if number % 2 == 0:
        return True if number == 2 else False
    middle = int(np.ceil(number / 2))
    for i in range(3, number, 2):
        if i > middle:
            break
        if number % i == 0:
            return False
    return True


from typing import (
    Any,
    ByteString,
    Dict,
    List,
    Optional,
    Union,
)
import io

import openpyxl
from openpyxl.utils import get_column_letter
from openpyxl.worksheet.dimensions import (
    ColumnDimension,
    DimensionHolder,
)
from openpyxl.worksheet.worksheet import Worksheet
import pandas as pd
import win32com.client

from helperpy.core.type_annotations import (
    Number,
    NumberOrString,
)


def style_dataframe_conditionally(
        df_style_obj: Any,
        props: List[Dict[str, Union[pd.Series, List[NumberOrString], Dict[str, str], None]]],
    ) -> Any:
    """
    Takes in Pandas Styler object (of class `pandas.io.formats.style.Styler`), and a list of `props`.
    Returns Pandas Styler object after applying the styles based on the given conditions (via `props`).
    Reference: https://pandas.pydata.org/pandas-docs/stable/user_guide/style.html
    
    In the `props`, each prop is a dictionary having 3 keys: ['row_conditions', 'column_subset', 'styling_kwargs'].
    The keys 'row_conditions' and 'column_subset' can have a value of None.
    The key 'styling_kwargs' can have a value of {}.
    The ordering of the `props` matter i.e; if 5 props are given, the 4th prop can over-write the 3rd prop.

    >>> num_rows = 50
    >>> df = pd.DataFrame(data={
        'column1': [np.random.randint(-25, 25) for _ in range(num_rows)],
        'column2': [np.random.randint(0, 100) for _ in range(num_rows)],
        'column3': [np.random.randint(-40, 15) for _ in range(num_rows)],
        'column4': [np.random.randint(-5, 5) for _ in range(num_rows)],
    })
    >>> props = [
        {
            'row_conditions': (df['column1'] >= 10) | (df['column1'] % 2 == 0),
            'column_subset': ['column1', 'column2'],
            'styling_kwargs': {
                'background-color': 'green',
                'color': 'purple',
            },
        },
        {
            'row_conditions': (df['column3'] < 0),
            'column_subset': None,
            'styling_kwargs': {
                'background-color': 'blue',
                'color': 'orange',
            },
        },
        {
            'row_conditions': None,
            'column_subset': ['column4'],
            'styling_kwargs': {
                'background-color': 'grey',
                'color': 'red',
            },
        },
    ]
    >>> style_dataframe_conditionally(df_style_obj=df.style, props=props)
    """
    df_styled = df_style_obj.__copy__()
    idx = pd.IndexSlice
    for prop in props:
        row_conditions = prop['row_conditions']
        column_subset = prop['column_subset']
        styling_kwargs = prop['styling_kwargs']

        subset = idx[idx[row_conditions], column_subset]
        if row_conditions is None and column_subset is None:
            subset = None
        elif row_conditions is None and column_subset is not None:
            subset = idx[idx[:], column_subset]
        elif row_conditions is not None and column_subset is None:
            subset = idx[idx[row_conditions], :]
        df_styled = df_styled.set_properties(
            subset=subset,
            axis=1,
            **styling_kwargs,
        )
    return df_styled


def style_dataframe_with_background_gradient(
        df_style_obj: Any,
        blue_white_red: Optional[List[NumberOrString]] = None,
        blues: Optional[List[NumberOrString]] = None,
        greens: Optional[List[NumberOrString]] = None,
        red_yellow_green: Optional[List[NumberOrString]] = None,
        summer: Optional[List[NumberOrString]] = None,
    ) -> Any:
    """
    Takes in Pandas Styler object (of class `pandas.io.formats.style.Styler`), and returns the same
    after applying background gradient styles based on the given column subsets.
    """
    df_styled = df_style_obj.__copy__()
    if blue_white_red:
        df_styled = df_styled.background_gradient(subset=blue_white_red, cmap='bwr')
    if blues:
        df_styled = df_styled.background_gradient(subset=blues, cmap='Blues')
    if greens:
        df_styled = df_styled.background_gradient(subset=greens, cmap='Greens')
    if red_yellow_green:
        df_styled = df_styled.background_gradient(subset=red_yellow_green, cmap='RdYlGn')
    if summer:
        df_styled = df_styled.background_gradient(subset=summer, cmap='summer')
    return df_styled


def save_styled_dataframes(
        destination_filepath: str,
        sheet_name_to_df_styled: Dict[str, Any],
    ) -> None:
    """
    Saves Pandas Styler object/s to an Excel file, which could have multiple sheets (one for each Styler object).
    >>> save_styled_dataframes(
        destination_filepath="resulting_file.xlsx",
        sheet_name_to_df_styled={
            "sheet1": df_style_obj_1,
            "sheet2": df_style_obj_2,
        },
    )
    """
    with pd.ExcelWriter(destination_filepath) as excel_writer:
        for sheet_name, df_styled in sheet_name_to_df_styled.items():
            df_styled.to_excel(
                excel_writer=excel_writer,
                sheet_name=sheet_name,
                index=False,
                engine='xlsxwriter',
            )
    return None


def get_all_excel_sheet_names(filepath: str) -> List[str]:
    """Returns list of all sheet names in an Excel file"""
    excel_file = pd.ExcelFile(filepath, engine='openpyxl')
    return excel_file.sheet_names


def get_all_excel_sheets_as_dataframes(filepath: str) -> Dict[str, pd.DataFrame]:
    """
    Takes `filepath` to an Excel file. Returns dictionary having keys = sheet names, and
    values = DataFrame corresponding to respective sheet name.
    """
    sheet_names = get_all_excel_sheet_names(filepath=filepath)
    dict_obj = {}
    for sheet_name in sheet_names:
        df_by_sheet = pd.read_excel(filepath, engine='openpyxl', sheet_name=sheet_name)
        dict_obj[sheet_name] = df_by_sheet
    return dict_obj


def excel_file_to_bytes(filepath: str) -> ByteString:
    """Takes `filepath` to an Excel file, and returns ByteString of the same"""
    bio = io.BytesIO()
    writer = pd.ExcelWriter(bio, engine='openpyxl')
    dict_all_excel_sheets = get_all_excel_sheets_as_dataframes(filepath=filepath)
    for sheet_name, df_obj in dict_all_excel_sheets.items():
        df_obj.to_excel(writer, sheet_name=sheet_name, index=False)
    writer.save()
    bytes_obj = bio.getvalue()
    return bytes_obj


def __get_column_names_from_worksheet(worksheet: Worksheet) -> List[NumberOrString]:
    """Returns list of column names (headers) from worksheet object (i.e; the values from row #1)"""
    column_names = [
        column.value for column in next(worksheet.iter_rows(min_row=1, max_row=1))
    ]
    return column_names


def adjust_column_widths_of_excel_file(
        src_filepath: str,
        destination_filepath: str,
        how: Optional[str] = 'column_header_and_values',
        fixed_width: Optional[Number] = None,
        sheets_subset: Optional[List[str]] = None,
    ) -> None:
    """
    Adjusts the widths of the columns in the Excel file given.
    Works only for non-hierarchical columns (flat columns).
    Note: Only works for Excel files with the 'xlsx' extension.

    Parameters:
        - src_filepath (str): Path to the source Excel file (along with the extension).
        - destination_filepath (str): Path to the destination Excel file (along with the extension).
        - how (str): How to assign the column widths. Options: ['column_header', 'column_header_and_values', 'fixed']. Default: 'column_header_and_values'.
        If `how` is set to 'column_header', the column widths will be the size of the column header.
        If `how` is set to 'column_header_and_values', the column widths will be the size of the longest of the column's header and values.
        If `how` is set to 'fixed', please set the `fixed_width` parameter.
        - fixed_width (int | float): Fixed width to use for the columns. To be set only if `how` is set to 'fixed'. Default: None.
        - sheets_subset (list): Subset of sheets for which the column widths need to be altered. Default: None.
    
    >>> adjust_column_widths_of_excel_file(
            src_filepath="data.xlsx",
            destination_filepath="data_with_adjusted_columns.xlsx",
            how='fixed',
            fixed_width=15,
            sheets_subset=['Sheet1', 'Sheet2', 'Sheet5'],
        )
    """
    how_options = ['column_header', 'column_header_and_values', 'fixed']
    if how not in how_options:
        raise ValueError(f"Expected `how` to be in {how_options}, but got '{how}'")
    if how == 'fixed' and fixed_width is None:
        raise ValueError("Since you have set how='fixed', please set the `fixed_width` parameter as well")
    
    workbook = openpyxl.load_workbook(filename=src_filepath)
    sheet_names = workbook.sheetnames if sheets_subset is None else sheets_subset
    for sheet_name in sheet_names:
        worksheet = workbook[sheet_name]
        column_names = __get_column_names_from_worksheet(worksheet=worksheet)
        dimension_holder = DimensionHolder(worksheet=worksheet)
        for column_idx, column_name in enumerate(column_names):
            column_idx += 1 # Since 1-based indexing is used
            column_letter = get_column_letter(idx=column_idx)
            if how == 'column_header':
                width = len(column_name) * 1.2
            elif how == 'column_header_and_values':
                cell_objs_by_column = worksheet[column_letter]
                width = max(map(lambda cell_obj: len(str(cell_obj.value)), cell_objs_by_column)) * 1.2
            elif how == 'fixed':
                width = fixed_width
            dimension_holder[column_letter] = ColumnDimension(
                worksheet=worksheet,
                min=column_idx,
                max=column_idx,
                width=width,
            )
        worksheet.column_dimensions = dimension_holder
    workbook.save(filename=destination_filepath)
    return None


def excel_to_pdf(
        src_filepath: str,
        destination_pdf_filepath: str,
        sheets_subset: Optional[List[str]] = None,
    ) -> None:
    r"""
    Takes `src_filepath` to an Excel file, and converts it to PDF file and saves it to `destination_pdf_filepath`.
    Note: Requires absolute filepaths. Does not work with relative filepaths.
    Reference: https://stackoverflow.com/questions/66421969/how-to-convert-excel-to-pdf-using-python

    >>> excel_to_pdf(
            src_filepath=r"C:\files\excel_files\some_data.xlsx",
            destination_pdf_filepath=r"C:\files\pdf_files\some_data.pdf",
            sheets_subset=['Sheet1', 'Sheet2', 'Sheet5'],
        )
    """
    sheet_names = get_all_excel_sheet_names(filepath=src_filepath)
    sheet_indices = [idx+1 for idx, _ in enumerate(sheet_names)] # Set 1-based indices
    if sheets_subset is not None:
        sheets_that_do_not_exist = list(
            set(sheets_subset).difference(set(sheet_names))
        )
        if sheets_that_do_not_exist:
            raise ValueError(f"The following sheet names do not exist: {sheets_that_do_not_exist}. Please correct the `sheets_subset` parameter")
        for sheet_idx, sheet_name in zip(sheet_indices, sheet_names):
            if sheet_name not in sheets_subset:
                sheet_indices.remove(sheet_idx)
                sheet_names.remove(sheet_name)
    dispatch_obj = win32com.client.Dispatch(dispatch="Excel.Application")
    dispatch_obj.Visible = False
    wb_obj = dispatch_obj.Workbooks.Open(src_filepath)
    wb_obj.WorkSheets(sheet_indices).Select()
    wb_obj.ActiveSheet.ExportAsFixedFormat(0, destination_pdf_filepath)
    wb_obj.Close()
    dispatch_obj.Quit()
    return None


import os
from setuptools import find_packages, setup

# Constants
PACKAGE_NAME = "helperpy"
PACKAGE_VERSION = "0.0.4"
AUTHOR_NAME = "Nishant Rao"
AUTHOR_EMAIL_ID = "nishant.rao173@gmail.com"
FILEPATH_TO_README = os.path.join(os.path.abspath(os.path.dirname(__file__)), "README.md")
FILEPATH_TO_REQUIREMENTS = os.path.join(os.path.abspath(os.path.dirname(__file__)), "requirements.txt")
REPOSITORY_URL = "https://github.com/Nishant173/helperpy"

# Requirements
install_requires = []
with open(file=FILEPATH_TO_REQUIREMENTS, mode="r") as fp:
    install_requires.extend(
        [s for s in [line.strip(" \n") for line in fp] if not s.startswith("#") and s != ""]
    )

# Setup
setup(
    name=PACKAGE_NAME,
    version=PACKAGE_VERSION,
    description="Library that serves a variety of utility functions (in Python)",
    long_description=f"Library that serves a variety of utility functions (in Python). Our repository: {REPOSITORY_URL}",
    author=AUTHOR_NAME,
    author_email=AUTHOR_EMAIL_ID,
    url=REPOSITORY_URL,
    packages=find_packages(where="."),
    include_package_data=True,
    install_requires=install_requires,
    license="MIT",
    classifiers=[
        "Development Status :: 1 - Planning",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: ISC License (ISCL)",
        "Natural Language :: English",
        "Programming Language :: Python :: 3.6",
        "Operating System :: Unix",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: Microsoft :: Windows",
    ],
)



from typing import List


def __raise_error_if_num_chars_is_negative(num_chars: int) -> None:
    if num_chars < 0:
        raise ValueError(f"Expected `num_chars` to be >= 0, but got {num_chars}")
    return None


def get_first_n_characters(text: str, num_chars: int) -> str:
    __raise_error_if_num_chars_is_negative(num_chars=num_chars)
    return text[:num_chars]


def get_last_n_characters(text: str, num_chars: int) -> str:
    __raise_error_if_num_chars_is_negative(num_chars=num_chars)
    return text[-num_chars:]


def remove_first_n_characters(text: str, num_chars: int) -> str:
    __raise_error_if_num_chars_is_negative(num_chars=num_chars)
    return text[num_chars:]


def remove_last_n_characters(text: str, num_chars: int) -> str:
    __raise_error_if_num_chars_is_negative(num_chars=num_chars)
    return text[:-num_chars]


def remove_characters_at_indices(text: str, indices: List[int]) -> str:
    """
    Removes characters present at the given `indices` in the `text`.
    Expects `indices` to be in range (0, n-1) where n is the length of the `text`.
    Raises an IndexError if any of the given `indices` are out of bounds.
    """
    if not indices:
        return text
    indices = sorted(list(set(indices)), reverse=True)
    lowest_possible_index, highest_possible_index = 0, len(text) - 1 # Must not use negative indices
    if indices[-1] < lowest_possible_index or indices[0] > highest_possible_index:
        raise IndexError(
            f"Accepted index-range for the given text is ({lowest_possible_index}, {highest_possible_index})."
            " Cannot remove character at an index outside of this range."
        )
    list_of_chars = list(text)
    for index in indices:
        list_of_chars.pop(index)
    return "".join(list_of_chars)


from typing import Dict, List, Optional

from datetime import datetime
import random
import string

from faker import Faker
import numpy as np
import pandas as pd

from helperpy.core.date_ops import get_random_timestamp
from helperpy.core.exceptions import raise_exception_if_invalid_option


def generate_random_hex_code() -> str:
    """Generates random 6-digit hexadecimal code"""
    choices = '0123456789ABCDEF'
    random_hex_code = '#'
    for _ in range(6):
        random_hex_code += random.choice(choices)
    return random_hex_code


def generate_random_hex_codes(how_many: int) -> List[str]:
    """Returns list of random 6-digit hexadecimal codes"""
    return [generate_random_hex_code() for _ in range(how_many)]


def generate_random_string(
        length: Optional[int] = 15,
        include_lowercase: Optional[bool] = True,
        include_uppercase: Optional[bool] = True,
        include_digits: Optional[bool] = True,
        include_punctuations: Optional[bool] = True,
    ) -> str:
    character_set = ""
    if include_lowercase:
        character_set += string.ascii_lowercase
    if include_uppercase:
        character_set += string.ascii_uppercase
    if include_digits:
        character_set += string.digits
    if include_punctuations:
        character_set += string.punctuation
    return "".join((random.choice(character_set) for _ in range(length)))


def generate_random_data(
        num_records: int,
        column_to_datatype_mapper: Dict[str, str],
        insert_random_nulls: Optional[bool] = False,
    ) -> pd.DataFrame:
    """
    Returns DataFrame having randomly generated fake data.

    Accepted data-type options for the columns in the DataFrame:
        - integer
        - float
        - string
        - date
        - timestamp
        - boolean
    
    >>> generate_random_data(
            num_records=1000,
            column_to_datatype_mapper={
                'name': 'string',
                'age': 'integer',
                'date_of_birth': 'date',
                'joined_at': 'timestamp',
                'salary': 'float',
                'is_recent_recruit': 'boolean',
            },
            insert_random_nulls=False,
        )
    """
    datatypes = list(column_to_datatype_mapper.values())
    for datatype in datatypes:
        raise_exception_if_invalid_option(
            option_name='datatype',
            option_value=datatype,
            valid_option_values=['integer', 'float', 'string', 'date', 'timestamp', 'boolean'],
        )
    dict_obj = {}
    for column, datatype in column_to_datatype_mapper.items():
        if datatype == 'integer':
            dict_obj[column] = (random.randint(-99999, 99999) for _ in range(num_records))
        elif datatype == 'float':
            dict_obj[column] = (random.random() * random.choice([10, 100, 1000, 10000, 100000]) * random.choice([-1, 1]) for _ in range(num_records))
        elif datatype == 'string':
            dict_obj[column] = (generate_random_string(include_punctuations=False) for _ in range(num_records))
        elif datatype == 'date':
            dict_obj[column] = (get_random_timestamp().date() for _ in range(num_records))
        elif datatype == 'timestamp':
            dict_obj[column] = (get_random_timestamp() for _ in range(num_records))
        elif datatype == 'boolean':
            dict_obj[column] = (random.choice([True, False]) for _ in range(num_records))
    df = pd.DataFrame(data=dict_obj)
    if insert_random_nulls:
        df.mask(cond=np.random.choice([True, False], size=df.shape, p=[0.2, 0.8]), inplace=True)
    return df


def generate_random_profile_data(num_records: int) -> pd.DataFrame:
    """
    Returns DataFrame having randomly generated fake data of people's profiles.
    Columns to be returned: ['job', 'company', 'ssn', 'residence', 'current_location', 'blood_group', 'website', 'username', 'name', 'sex', 'address', 'mail', 'birthdate', 'birthdatetime']

    Reference:
        - https://towardsdatascience.com/generating-fake-data-with-python-c7a32c631b2a
        - https://www.caktusgroup.com/blog/2020/04/15/quick-guide-generating-fake-data-with-pandas/
    """
    faker = Faker()
    df = pd.DataFrame(data=(faker.profile() for _ in range(num_records)))
    df['birthdatetime'] = df['birthdate'].apply(
        lambda date_obj: datetime(
            year=date_obj.year,
            month=date_obj.month,
            day=date_obj.day,
            hour=random.choice(range(0, 23+1)),
            minute=random.choice(range(0, 59+1)),
            second=random.choice(range(0, 59+1)),
        )
    )
    return df


from typing import Any, Dict, List, Optional, Tuple
import random

import numpy as np

from helperpy.core.exceptions import (
    raise_exception_if_invalid_option,
    raise_exception_if_invalid_type,
)
from helperpy.core.type_annotations import (
    Number,
    NumberOrString,
)


def print_docstring(obj: Any) -> None:
    """Prints the doc-string (if available). Usually of a class, method or function."""
    if hasattr(obj, "__doc__"):
        print(obj.__doc__)
    return None


def class_object_to_string_repr(
        class_name: str,
        details: Dict[str, Any],
    ) -> str:
    """
    Writes __str__ representation of a class' instance.

    >>> class_object_to_string_repr(
            class_name="Person",
            details={
                "first_name": "James",
                "last_name": "Murphy",
                "age": 35,
                "is_developer": True,
            },
        ) # Returns the string: "Person(first_name='James', last_name='Murphy', age=35, is_developer=True)"
    """
    kwargs_as_strings = []
    for key, value in details.items():
        value_as_string = f"'{value}'" if isinstance(value, str) else str(value)
        kwargs_as_strings.append(f"{key}={value_as_string}")
    return f"{class_name}({', '.join(kwargs_as_strings)})"


def is_none(value: Any) -> bool:
    return value is None


def is_nan(value: Any) -> bool:
    """Returns True if value given is Numpy's NaN. Otherwise returns False"""
    if isinstance(value, float):
        if np.isnan(value):
            return True
    return False


def is_none_or_nan(value: Any) -> bool:
    """Returns True if value given is Python's native None or Numpy's NaN. Otherwise returns False"""
    return is_none(value=value) or is_nan(value=value)


def get_timetaken_dictionary(num_seconds: Number) -> Dict[str, Number]:
    """
    Returns dictionary having the keys: ['weeks', 'days', 'hours', 'minutes', 'seconds', 'milliseconds'] containing the time elapsed.
    
    >>> get_timetaken_dictionary(num_seconds=3725.4292)
    >>> get_timetaken_dictionary(num_seconds=885354.128129)
    """
    weeks, remainder = divmod(num_seconds, 60*60*24*7)
    days, remainder = divmod(remainder, 60*60*24)
    hours, remainder = divmod(remainder, 60*60)
    minutes, remainder = divmod(remainder, 60)
    seconds = np.floor(remainder)
    milliseconds = round((remainder - seconds) * 1000, 5)
    
    dictionary_time_taken = {
        'weeks': integerify_if_possible(weeks),
        'days': integerify_if_possible(days),
        'hours': integerify_if_possible(hours),
        'minutes': integerify_if_possible(minutes),
        'seconds': integerify_if_possible(seconds),
        'milliseconds': integerify_if_possible(milliseconds),
    }
    return dictionary_time_taken


def get_timetaken_fstring(num_seconds: Number) -> str:
    """Returns f-string containing the elapsed time"""
    dict_time_taken = get_timetaken_dictionary(num_seconds=num_seconds)
    dict_unit_shortener = {
        'weeks': 'w',
        'days': 'd',
        'hours': 'h',
        'minutes': 'm',
        'seconds': 's',
        'milliseconds': 'ms',
    }
    time_taken_fstring = " ".join([f"{value}{dict_unit_shortener.get(unit, unit)}" for unit, value in dict_time_taken.items() if value != 0]).strip()
    time_taken_fstring = '0s' if time_taken_fstring == "" else time_taken_fstring
    return time_taken_fstring


def get_random_choice_except(
        choices: List[NumberOrString],
        exception: NumberOrString,
    ) -> NumberOrString:
    """Gets random choice from an iterable, given an exception"""
    if exception not in choices:
        raise ValueError("The `exception` is not available in the given `choices`")
    choices_available = list(set(choices).difference(set([exception])))
    if not choices_available:
        raise ValueError("No choices available")
    return random.choice(choices_available)


def round_off_as_string(number: Number, round_by: int) -> str:
    """
    Rounds off the given `number` to `round_by` decimal places, and type casts
    it to a string (to retain the exact number of decimal places desired).
    """
    if round_by < 0:
        raise ValueError("The `round_by` parameter must be >= 0")
    if round_by == 0:
        return str(round(number))
    number_stringified = str(round(number, round_by))
    decimal_places_filled = len(number_stringified.split('.')[-1])
    decimal_places_to_fill = round_by - decimal_places_filled
    for _ in range(decimal_places_to_fill):
        number_stringified += '0'
    return number_stringified


def commafy_number(number: Number) -> str:
    """
    Adds commas to number for better readability.
    >>> commafy_number(number=1738183090) # Returns "1,738,183,090"
    >>> commafy_number(number=1738183090.90406) # Returns "1,738,183,090.90406"
    """
    if int(number) == number:
        return format(int(number), ",d")
    return format(number, ",f")


def is_whole_number(number: Number) -> bool:
    return int(number) == number


def integerify_if_possible(number: Number) -> Number:
    """Converts whole numbers represented as floats to integers"""
    return int(number) if is_whole_number(number=number) else number


def string_to_int_or_float(value: str) -> Number:
    """Converts stringified number to either int or float"""
    number = float(value)
    number = integerify_if_possible(number=number)
    return number


def stringify_list_of_nums(array: List[Number]) -> str:
    """Converts list of ints/floats to comma separated string of the same"""
    return ",".join(list(map(str, array)))


def listify_string_of_nums(string: str) -> List[Number]:
    """Converts string of comma separated ints/floats to list of numbers"""
    numbers = string.split(',')
    numbers = list(map(string_to_int_or_float, numbers))
    return numbers


def get_max_of_abs_values(array: List[Number]) -> Number:
    """Finds maximum of absolute values of numbers in an array"""
    array_abs = list(map(abs, array))
    return max(array_abs)


def get_min_of_abs_values(array: List[Number]) -> Number:
    """Finds minimum of absolute values of numbers in an array"""
    array_abs = list(map(abs, array))
    return min(array_abs)


def has_negative_number(array: List[Number]) -> bool:
    for number in array:
        if number < 0:
            return True
    return False


def has_positive_number(array: List[Number]) -> bool:
    for number in array:
        if number > 0:
            return True
    return False


def cumulative_aggregate(numbers: List[Number], method: str) -> List[Number]:
    """
    Returns list of cumulative aggregates.
    Options for `method` are: ['sum', 'difference', 'product', 'division'].
    """
    raise_exception_if_invalid_option(
        option_name='method',
        option_value=method,
        valid_option_values=['sum', 'difference', 'product', 'division'],
    )

    length = len(numbers)
    if length == 0:
        return []
    cumulative_array = [numbers[0]]
    if length == 1:
        return cumulative_array
    method_mapper = {
        'sum': lambda x, y: x + y,
        'difference': lambda x, y: x - y,
        'product': lambda x, y: x * y,
        'division': lambda x, y: x / y,
    }
    for number in numbers[1:]:
        cumulative_array.append(method_mapper[method](cumulative_array[-1], number))
    return cumulative_array



class Partitioner:
    """
    Class for iterable partitioning.

    >>> partitioner = Partitioner(iterable_length=100)
    >>> kwargs = {
            'num_partitions': 8,
            'distribution_method': 'uniform',
        }
    >>> partitioner.sizes_by_num_partitions(**kwargs)
    >>> partitioner.index_ranges_by_num_partitions(**kwargs)
    >>> partitioner.sizes_by_max_partition_size(max_partition_size=15)
    >>> partitioner.index_ranges_by_max_partition_size(max_partition_size=15)
    """

    def __init__(self, iterable_length: int) -> None:
        self.__validate_iterable_length(iterable_length=iterable_length)
        self.__iterable_length = iterable_length
    
    def __str__(self) -> str:
        return f"{self.__class__.__name__}(iterable_length={self.__iterable_length})"
    
    def __validate_iterable_length(self, iterable_length: int) -> None:
        raise_exception_if_invalid_type(
            parameter_name='iterable_length',
            parameter_value=iterable_length,
            expected_type=int,
        )
        if iterable_length < 0:
            raise ValueError("The parameter `iterable_length` cannot be < 0")
        return None
    
    def __validate_max_partition_size(self, max_partition_size: int) -> None:
        raise_exception_if_invalid_type(
            parameter_name='max_partition_size',
            parameter_value=max_partition_size,
            expected_type=int,
        )
        if max_partition_size <= 0:
            raise ValueError("The parameter `max_partition_size` cannot be <= 0")
        if max_partition_size > self.__iterable_length:
            raise ValueError("The parameter `max_partition_size` cannot be > length of the iterable")
        return None
    
    def __validate_num_partitions(self, num_partitions: int) -> None:
        raise_exception_if_invalid_type(
            parameter_name='num_partitions',
            parameter_value=num_partitions,
            expected_type=int,
        )
        if num_partitions <= 0:
            raise ValueError("The parameter `num_partitions` cannot be <= 0")
        if num_partitions > self.__iterable_length:
            raise ValueError("The parameter `num_partitions` cannot be > length of the iterable")
        return None
    
    def sizes_by_max_partition_size(self, max_partition_size: int) -> List[int]:
        """Returns list of sizes of each partition"""
        self.__validate_max_partition_size(max_partition_size=max_partition_size)
        num_full_partitions, last_partition_size = divmod(self.__iterable_length, max_partition_size)
        if last_partition_size == 0:
            return [max_partition_size] * num_full_partitions
        return [max_partition_size] * num_full_partitions + [last_partition_size]
    
    def __sizes_by_fill_distribution(self, num_partitions: int) -> List[int]:
        full_partition_size, last_partition_size = divmod(self.__iterable_length, num_partitions)
        if last_partition_size == 0:
            return [full_partition_size] * num_partitions
        sizes = [full_partition_size] * (num_partitions - 1)
        sizes.insert(0, full_partition_size + last_partition_size)
        return sizes
    
    def __sizes_by_uniform_distribution(self, num_partitions: int) -> List[int]:
        min_partition_size, num_residuals = divmod(self.__iterable_length, num_partitions)
        return [min_partition_size + 1] * num_residuals + [min_partition_size] * (num_partitions - num_residuals)
    
    def sizes_by_num_partitions(
            self,
            num_partitions: int,
            distribution_method: Optional[str] = 'uniform',
        ) -> List[int]:
        """
        Returns list of sizes of each partition.
        Options for `distribution_method` are: ['fill', 'uniform']. Default: 'uniform'
        """
        self.__validate_num_partitions(num_partitions=num_partitions)
        raise_exception_if_invalid_option(
            option_name='distribution_method',
            option_value=distribution_method,
            valid_option_values=['fill', 'uniform'],
        )
        if distribution_method == 'fill':
            return self.__sizes_by_fill_distribution(num_partitions=num_partitions)
        if distribution_method == 'uniform':
            return self.__sizes_by_uniform_distribution(num_partitions=num_partitions)
    
    def __compute_index_ranges_from_partition_sizes(
            self,
            partition_sizes: List[int],
        ) -> List[Tuple[int, int]]:
        start_indices = [0] + list(np.cumsum(partition_sizes))
        index_ranges = [(start_indices[idx], start_indices[idx+1]) for idx in range(len(start_indices) - 1)]
        return index_ranges
    
    def index_ranges_by_max_partition_size(self, max_partition_size: int) -> List[Tuple[int, int]]:
        """
        Returns list of tuples of (start_index, end_index) that partition an iterable.
        The indices are zero-based.
        """
        sizes = self.sizes_by_max_partition_size(max_partition_size=max_partition_size)
        return self.__compute_index_ranges_from_partition_sizes(partition_sizes=sizes)
    
    def index_ranges_by_num_partitions(
            self,
            num_partitions: int,
            distribution_method: Optional[str] = 'uniform',
        ) -> List[Tuple[int, int]]:
        """
        Returns list of tuples of (start_index, end_index) that partition an iterable.
        The indices are zero-based.
        Options for `distribution_method` are: ['fill', 'uniform']. Default: 'uniform'
        """
        sizes = self.sizes_by_num_partitions(num_partitions=num_partitions, distribution_method=distribution_method)
        return self.__compute_index_ranges_from_partition_sizes(partition_sizes=sizes)